#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
This dialogue is used to configure different log configurations that is used to
enable logging of data from the Crazyflie. These can then be used in different
views in the UI.
"""

import logging
import struct

import cfclient
from cfclient.utils.ui import UiUtils
from PyQt6 import QtWidgets, uic
from PyQt6.QtCore import Qt, QTimer
from PyQt6.QtGui import QShortcut, QKeySequence

from cflib.crazyflie.log import LogConfig

__author__ = 'Bitcraze AB'
__all__ = ['LogConfigDialogue']

logger = logging.getLogger(__name__)

(logconfig_widget_class, connect_widget_base_class) = (
    uic.loadUiType(cfclient.module_path + '/ui/dialogs/logconfigdialogue.ui'))

NAME_FIELD = 0
ID_FIELD = 1
TYPE_FIELD = 2
SIZE_FIELD = 3
MAX_LOG_SIZE = 26


class LogConfigDialogue(QtWidgets.QWidget, logconfig_widget_class):

    def __init__(self, helper, *args):
        super(LogConfigDialogue, self).__init__(*args)
        self.setupUi(self)
        self.helper = helper

        self.logTree.setHeaderLabels(['Name', 'ID', 'Type', 'Size'])
        self.varTree.setHeaderLabels(['Name', 'ID', 'Type', 'Size'])
        self.categoryTree.setHeaderLabels(['Categories'])

        self.logTree.setSortingEnabled(True)
        self.varTree.setSortingEnabled(True)

        # Item-click callbacks.
        self.addButton.clicked.connect(lambda: self.moveNode(self.logTree,
                                                             self.varTree))
        self.removeButton.clicked.connect(lambda: self.moveNode(self.varTree,
                                                                self.logTree))
        self.saveButton.clicked.connect(self.saveConfig)

        self.categoryTree.itemSelectionChanged.connect(self._item_selected)
        self.categoryTree.itemPressed.connect(self._on_item_press)
        self.categoryTree.itemChanged.connect(self._config_changed)

        # Add/remove item on doubleclick.
        self.logTree.itemDoubleClicked.connect(self.itemDoubleClicked)
        self.varTree.itemDoubleClicked.connect(lambda: self.moveNode(
                                            self.varTree, self.logTree))
        self.loggingPeriod.textChanged.connect(self.periodChanged)

        self.currentSize = 0
        self.packetSize.setMaximum(100)
        self.packetSize.setValue(0)
        self.period = 0

        # Used when renaming a config/category
        self._last_pressed_item = None

        # set icons
        save_icon, delete_icon = self.helper.logConfigReader.get_icons()
        self.createCategoryBtn.setIcon(save_icon)
        self.createConfigBtn.setIcon(save_icon)
        self.deleteBtn.setIcon(delete_icon)

        # bind buttons
        self.createCategoryBtn.clicked.connect(self._create_category)
        self.createConfigBtn.clicked.connect(self._create_config)
        self.deleteBtn.clicked.connect(self._delete_config)

        # set tooltips
        self.createCategoryBtn.setToolTip('Create a new category')
        self.createConfigBtn.setToolTip('Create a new log-config')
        self.deleteBtn.setToolTip('Delete category')

        # enable right-click context-menu
        self.categoryTree.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.categoryTree.customContextMenuRequested.connect(
                                                    self.menuContextTree)

        # keyboard shortcuts
        shortcut_delete = QShortcut(QKeySequence("Delete"), self)
        shortcut_delete.activated.connect(self._delete_config)

        shortcut_f2 = QShortcut(QKeySequence("F2"), self)
        shortcut_f2.activated.connect(self._edit_name)

        self._config_saved_timer = QTimer()
        self._config_saved_timer.timeout.connect(self._config_saved_status)

        self.closeOnSave.setChecked(True)

    def itemDoubleClicked(self):
        if self.categoryTree.selectedItems():
            self.moveNode(self.logTree, self.varTree)

    def _config_saved_status(self):
        self.statusText.setText('')
        self._config_saved_timer.stop()

    def _on_item_press(self, item):
        self._last_pressed_item = item, item.text(0)

    def _create_config(self):
        """ Creates a new log-configuration in the chosen
            category. If no category is selected, the
            configuration is stored in the 'Default' category.
        """
        items = self.categoryTree.selectedItems()

        if items:
            config = items[0]
            parent = config.parent()
            if parent:
                category = parent.text(0)
            else:
                category = config.text(0)

            conf_name = self.helper.logConfigReader.create_empty_log_conf(
                                                category)
            self._reload()
            # Load the newly created log-config.
            self._select_item(conf_name, category)
            self._edit_name()

    def _create_category(self):
        """ Creates a new category and enables editing the name.  """
        category_name = self.helper.logConfigReader.create_category()
        self._load_saved_configs()
        self.sortTrees()
        self._select_category(category_name)
        self._edit_name()

    def _delete_config(self):

        """ Deletes a category or a configuration
            depending on if the item has a parent or not.
        """

        items = self.categoryTree.selectedItems()
        if items:
            config = items[0]
            parent = config.parent()

            if parent:
                # Delete a configuration in the given category.
                category = parent.text(0)
                self.helper.logConfigReader.delete_config(config.text(0),
                                                          category)
                self._reload()
            else:
                # Delete a category and all its log-configurations
                category = config.text(0)
                if category != 'Default':
                    self.helper.logConfigReader.delete_category(category)
                    self._reload()

    def _config_changed(self, config):
        """ Changes the name for a log-configuration or a category.
            This is a callback function that gets called when an item
            is changed.
        """
        item, old_name = self._last_pressed_item

        parent = config.parent()
        if parent:
            # Change name for a log-config, inside of the category.
            new_conf_name = item.text(0)
            category = parent.text(0)
            self.helper.logConfigReader.change_name_config(old_name,
                                                           new_conf_name,
                                                           category)
        else:
            # Change name for the category.
            category = config.text(0)
            self.helper.logConfigReader.change_name_category(old_name,
                                                             category)

    def _edit_name(self):
        """ Enables editing the clicked item.
            When the edit is saved, a callback is fired.
        """
        items = self.categoryTree.selectedItems()
        if items:
            item_clicked = items[0]
            self.categoryTree.editItem(item_clicked, 0)

    def _reload(self):
        self.resetTrees()
        self._load_saved_configs()
        self.sortTrees()

    def menuContextTree(self, point):

        menu = QtWidgets.QMenu()

        createConfig = None
        createCategory = None
        delete = None
        edit = None

        item = self.categoryTree.itemAt(point)
        if item:
            createConfig = menu.addAction('Create new log configuration')
            edit = menu.addAction('Edit name')

            if item.parent():
                delete = menu.addAction('Delete config')
            else:
                delete = menu.addAction('Delete category')
        else:
            createCategory = menu.addAction('Create new Category')

        action = menu.exec_(self.categoryTree.mapToGlobal(point))

        if action == createConfig:
            self._create_config()
        elif createCategory:
            self._create_category()
        elif action == delete:
            self._delete_config()
        elif action == edit:
            self._edit_name()

    def _select_category(self, category):
        items = self.categoryTree.findItems(category, Qt.MatchFlag.MatchFixedString | Qt.MatchFlag.MatchRecursive)
        if items:
            category = items[0]
            self.categoryTree.setCurrentItem(category)
            self._last_pressed_item = category, category.text(0)

    def _select_item(self, conf_name, category):
        """ loads the given config in the correct category """
        items = self.categoryTree.findItems(conf_name, Qt.MatchFlag.MatchFixedString | Qt.MatchFlag.MatchRecursive)
        for item in items:
            if item.parent().text(0) == category:
                self._last_pressed_item = item, conf_name
                self._loadConfig(category, conf_name)
                self.categoryTree.setCurrentItem(item)

    def _item_selected(self):
        """ Opens the log configuration of the pressed
            item in the category-tree. """
        items = self.categoryTree.selectedItems()

        if items:
            config = items[0]
            category = config.parent()
            if category:
                self._loadConfig(category.text(NAME_FIELD),
                                 config.text(NAME_FIELD))
            else:
                # if category is None, it's the category that's clicked
                self._clear_trees_and_progressbar()

    def _clear_trees_and_progressbar(self):
        self.varTree.clear()
        self.logTree.clear()
        self.currentSize = 0
        self.loggingPeriod.setText('')
        self.updatePacketSizeBar()

    def _load_saved_configs(self):
        """ Read saved log-configs and display them on
            the left-side category-tree. """

        config = None
        config = self.helper.logConfigReader._getLogConfigs()

        if (config is None):
            logger.warning("Could not load config")
        else:
            self.categoryTree.clear()
            # Create category-tree.
            for conf_category in config:
                category = QtWidgets.QTreeWidgetItem()
                category.setData(NAME_FIELD, Qt.ItemDataRole.DisplayRole, conf_category)
                category.setFlags(category.flags() | Qt.ItemFlag.ItemIsEditable)

                # Copulate category-tree with log configurations.
                for conf in config[conf_category]:
                    item = QtWidgets.QTreeWidgetItem()

                    # Check if name contains category/config-name.
                    # This is only true is a new config has been added
                    # during a session, and the window re-opened.
                    if '/' in conf.name:
                        conf_name = conf.name.split('/')[1]
                    else:
                        conf_name = conf.name

                    item.setData(NAME_FIELD, Qt.ItemDataRole.DisplayRole, conf_name)
                    category.addChild(item)

                    # Enable item-editing.
                    item.setFlags(item.flags() | Qt.ItemFlag.ItemIsEditable)

                self.categoryTree.addTopLevelItem(category)
                self.categoryTree.expandItem(category)

            self.sortTrees()

    def _loadConfig(self, category, config_name):
        configs = self.helper.logConfigReader._getLogConfigs()[category]

        if (configs is None):
            logger.warning("Could not load config")

        else:
            for config in configs:
                name = self._parse_configname(config)
                if name == config_name:
                    self.resetTrees()
                    self.loggingPeriod.setText("%d" % config.period_in_ms)
                    self.period = config.period_in_ms
                    for v in config.variables:
                        if (v.is_toc_variable()):
                            parts = v.name.split(".")
                            varParent = parts[0]
                            varName = parts[1]
                            if self.moveNodeByName(
                                    self.logTree, self.varTree, varParent,
                                    varName) is False:
                                logger.warning("Could not find node %s.%s!!",
                                               varParent, varName)
                        else:
                            logger.warning("Error: Mem vars not supported!")

            self.sortTrees()

    def resetTrees(self):
        self.varTree.clear()
        self.logTree.clear()
        self.updateToc()

    def sortTrees(self):
        """ Sorts all trees by their name. """
        for tree in [self.logTree, self.varTree, self.categoryTree]:
            tree.sortItems(NAME_FIELD, Qt.SortOrder.AscendingOrder)
            for node in self.getNodeChildren(tree.invisibleRootItem()):
                node.sortChildren(NAME_FIELD, Qt.SortOrder.AscendingOrder)

    def getNodeChildren(self, treeNode):
        children = []
        for i in range(treeNode.childCount()):
            children.append(treeNode.child(i))
        return children

    def updatePacketSizeBar(self):
        self.currentSize = 0
        for node in self.getNodeChildren(self.varTree.invisibleRootItem()):
            for leaf in self.getNodeChildren(node):
                self.currentSize = (self.currentSize +
                                    int(leaf.text(SIZE_FIELD)))

        self.packetSizeText.setText('%s/%s bytes' % (self.currentSize,
                                                     MAX_LOG_SIZE))

        if self.currentSize > MAX_LOG_SIZE:
            self.packetSize.setMaximum(self.currentSize / MAX_LOG_SIZE * 100)
            self.packetSize.setFormat("%v%")
            self.packetSize.setValue(int(self.currentSize / MAX_LOG_SIZE * 100))
            self.packetSize.setStyleSheet(
                        UiUtils.progressbar_stylesheet('red'))
        else:
            self.packetSize.setMaximum(MAX_LOG_SIZE)
            self.packetSize.setFormat("%p%")
            self.packetSize.setValue(int(self.currentSize))
            self.packetSize.setStyleSheet(
                        UiUtils.progressbar_stylesheet(UiUtils.COLOR_GREEN))

    def addNewVar(self, logTreeItem, target):
        parentName = logTreeItem.parent().text(NAME_FIELD)
        varParent = target.findItems(parentName, Qt.MatchFlag.MatchExactly, NAME_FIELD)

        item = logTreeItem.clone()

        if (len(varParent) == 0):
            newParent = QtWidgets.QTreeWidgetItem()
            newParent.setData(0, Qt.ItemDataRole.DisplayRole, parentName)
            newParent.addChild(item)
            target.addTopLevelItem(newParent)
            target.expandItem(newParent)
        else:
            parent = varParent[0]
            parent.addChild(item)

    def moveNodeItem(self, source, target, item):
        if (item.parent() is None):
            children = self.getNodeChildren(item)
            for c in children:
                self.addNewVar(c, target)
            source.takeTopLevelItem(source.indexOfTopLevelItem(item))
        elif (item.parent().childCount() > 1):
            self.addNewVar(item, target)
            item.parent().removeChild(item)
        else:
            self.addNewVar(item, target)
            # item.parent().removeChild(item)
            source.takeTopLevelItem(source.indexOfTopLevelItem(item.parent()))
        self.updatePacketSizeBar()
        self.sortTrees()
        self.checkAndEnableSaveButton()

    def checkAndEnableSaveButton(self):
        if self.currentSize > 0 and self.period > 0 and self.currentSize <= 26:
            self.saveButton.setEnabled(True)
        else:
            self.saveButton.setEnabled(False)

    def moveNode(self, source, target):
        self.moveNodeItem(source, target, source.currentItem())

    def moveNodeByName(self, source, target, parentName, itemName):
        parents = source.findItems(parentName, Qt.MatchFlag.MatchExactly, NAME_FIELD)
        node = None
        if (len(parents) > 0):
            parent = parents[0]
            for n in range(parent.childCount()):
                if (parent.child(n).text(NAME_FIELD) == itemName):
                    node = parent.child(n)
                    break
        if (node is not None):
            self.moveNodeItem(source, target, node)
            return True
        return False

    def showEvent(self, event):
        self._clear_trees_and_progressbar()
        self._load_saved_configs()

    def periodChanged(self, value):
        try:
            self.period = int(value)
            self.checkAndEnableSaveButton()
        except Exception:
            self.period = 0

    def showErrorPopup(self, caption, message):
        self.box = QtWidgets.QMessageBox()  # noqa
        self.box.setWindowTitle(caption)
        self.box.setText(message)
        # self.box.setButtonText(1, "Ok")
        self.box.setWindowFlags(Qt.WindowType.Dialog | Qt.WindowType.MSWindowsFixedSizeDialogHint)
        self.box.show()

    def updateToc(self):
        self.logTree.clear()
        toc = self.helper.cf.log.toc

        for group in list(toc.toc.keys()):
            groupItem = QtWidgets.QTreeWidgetItem()
            groupItem.setData(NAME_FIELD, Qt.ItemDataRole.DisplayRole, group)
            for param in list(toc.toc[group].keys()):
                item = QtWidgets.QTreeWidgetItem()
                item.setData(NAME_FIELD, Qt.ItemDataRole.DisplayRole, param)
                item.setData(ID_FIELD, Qt.ItemDataRole.DisplayRole,
                             toc.toc[group][param].ident)
                item.setData(TYPE_FIELD, Qt.ItemDataRole.DisplayRole,
                             toc.toc[group][param].ctype)
                item.setData(SIZE_FIELD, Qt.ItemDataRole.DisplayRole,
                             struct.calcsize(toc.toc[group][param].pytype))
                groupItem.addChild(item)

            self.logTree.addTopLevelItem(groupItem)

    def populateDropDown(self):
        self.configNameCombo.clear()
        toc = self.helper.logConfigReader.getLogConfigs()
        for d in toc:
            self.configNameCombo.addItem(d.name)
        if (len(toc) > 0):
            self.loadButton.setEnabled(True)

    def saveConfig(self):
        items = self.categoryTree.selectedItems()

        if items:
            config = items[0]
            parent = config.parent()

            if parent:

                # If we're just editing an existing config, we'll delete
                # the old one first.
                self._delete_from_plottab(self._last_pressed_item[1])

                category = parent.text(NAME_FIELD)
                config_name = config.text(NAME_FIELD)
                updatedConfig = self.createConfigFromSelection(config_name)

                if category != 'Default':
                    plot_tab_name = '%s/%s' % (category, config_name)
                else:
                    plot_tab_name = config_name

                try:
                    self.helper.logConfigReader.saveLogConfigFile(
                                                            category,
                                                            updatedConfig)
                    self.statusText.setText('Log config succesfully saved!')
                    self._config_saved_timer.start(4000)
                    if self.closeOnSave.isChecked():
                        self.close()

                except Exception as e:
                    self.showErrorPopup("Error when saving file",
                                        "Error: %s" % e)

        # The name of the config is changed due to displaying
        # it as category/config-name in the plotter-tab.
        # The config is however saved with only the config-name.
        updatedConfig.name = plot_tab_name
        self.helper.cf.log.add_config(updatedConfig)

    def _parse_configname(self, config):
        """ If the configs are placed in a category,
            they are named as Category/confname.
        """
        parts = config.name.split('/')
        return parts[1] if len(parts) > 1 else parts[0]

    def _delete_from_plottab(self, conf_name):
        """ Removes a config from the plot-tab. """
        for logconfig in self.helper.cf.log.log_blocks:
            config_to_delete = self._parse_configname(logconfig)
            if config_to_delete == conf_name:
                self.helper.plotTab.remove_config(logconfig)
                self.helper.cf.log.log_blocks.remove(logconfig)
                logconfig.delete()

    def _get_node_children(self):
        root_item = self.varTree.invisibleRootItem()
        return [root_item.child(i) for i in range(root_item.childCount())]

    def createConfigFromSelection(self, config):
        logconfig = LogConfig(config, self.period)

        for node in self._get_node_children():
            parentName = node.text(NAME_FIELD)

            for leaf in self.getNodeChildren(node):
                varName = leaf.text(NAME_FIELD)
                varType = str(leaf.text(TYPE_FIELD))
                completeName = "%s.%s" % (parentName, varName)
                logconfig.add_variable(completeName, varType)

        return logconfig



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
#  USA.
"""
Driver for reading data from the PySDL2 API. Used from Inpyt.py for reading
input data.
"""
import logging
import sys
import time
from threading import Thread

from queue import Queue

if sys.platform.startswith('linux'):
    raise Exception("No SDL2 support on Linux")

try:
    import sdl2
    import sdl2.ext
    import sdl2.hints
except ImportError as e:
    raise Exception("sdl2 library probably not installed ({})".format(e))

__author__ = 'Bitcraze AB'
__all__ = ['PySDL2Reader']

logger = logging.getLogger(__name__)

MODULE_MAIN = "PySDL2Reader"
MODULE_NAME = "PySDL2"


class _SDLEventDispatcher(Thread):
    """Wrapper to read all SDL2 events from the global queue and distribute
    them to the different devices"""

    def __init__(self, callback):
        Thread.__init__(self)
        self._callback = callback
        self.daemon = True
        # SDL2 will Seg Fault on Linux if you read events after you
        # have closed a device (and without opening a new one). Even if you
        # have two devices open, it will crash after one.
        self.enable = False

    def run(self):
        while True:
            if self.enable:
                for ev in sdl2.ext.get_events():
                    try:
                        if self._callback:
                            self._callback(ev.jdevice.which, ev)
                    except AttributeError:
                        pass
            time.sleep(0.01)


class _JS():
    """Wrapper for one input device"""

    def __init__(self, sdl_index, sdl_id, name):
        self.axes = []
        self.buttons = []
        self.name = MODULE_NAME
        self._j = None
        self._id = sdl_id
        self._index = sdl_index
        self._name = name
        self._event_queue = Queue()

    def open(self):
        self._j = sdl2.SDL_JoystickOpen(self._index)
        btn_count = sdl2.SDL_JoystickNumButtons(self._j)

        self.axes = list(0 for i in range(sdl2.SDL_JoystickNumAxes(self._j)))
        self.buttons = list(0 for i in range(btn_count + 4))

    def close(self):
        if self._j:
            sdl2.joystick.SDL_JoystickClose(self._j)
        self._j = None

    def _set_virtual_dpad(self, position):
        self.buttons[-4:] = [0, 0, 0, 0]

        # -4 UP -3 DOWN -2 LEFT -1 RIGHT
        if position == sdl2.SDL_HAT_UP:
            self.buttons[-4] = 1
        elif position == sdl2.SDL_HAT_RIGHTUP:
            self.buttons[-4] = 1
            self.buttons[-1] = 1
        elif position == sdl2.SDL_HAT_RIGHT:
            self.buttons[-1] = 1
        elif position == sdl2.SDL_HAT_RIGHTDOWN:
            self.buttons[-1] = 1
            self.buttons[-3] = 1
        elif position == sdl2.SDL_HAT_DOWN:
            self.buttons[-3] = 1
        elif position == sdl2.SDL_HAT_LEFTDOWN:
            self.buttons[-3] = 1
            self.buttons[-2] = 1
        elif position == sdl2.SDL_HAT_LEFT:
            self.buttons[-2] = 1
        elif position == sdl2.SDL_HAT_LEFTUP:
            self.buttons[-2] = 1
            self.buttons[-4] = 1

    def add_event(self, event):
        self._event_queue.put(event)

    def read(self):
        while not self._event_queue.empty():
            e = self._event_queue.get_nowait()
            if e.type == sdl2.SDL_JOYAXISMOTION:
                self.axes[e.jaxis.axis] = e.jaxis.value / 32767.0

            if e.type == sdl2.SDL_JOYBUTTONDOWN:
                self.buttons[e.jbutton.button] = 1

            if e.type == sdl2.SDL_JOYBUTTONUP:
                self.buttons[e.jbutton.button] = 0

            if e.type == sdl2.SDL_JOYHATMOTION:
                self._set_virtual_dpad(e.jhat.value)

        return [self.axes, self.buttons]


class PySDL2Reader():
    """Used for reading data from input devices using the PySDL2 API."""

    def __init__(self):
        sdl2.SDL_Init(sdl2.SDL_INIT_JOYSTICK)
        sdl2.SDL_SetHint(sdl2.hints.SDL_HINT_JOYSTICK_ALLOW_BACKGROUND_EVENTS,
                         b"1")
        self._js = {}
        self.name = MODULE_NAME
        self._event_dispatcher = _SDLEventDispatcher(self._dispatch_events)
        self._event_dispatcher.start()
        self._devices = []

    def open(self, device_id):
        """Initialize the reading and open the device with deviceId and set
        the mapping for axis/buttons using the inputMap"""
        self._event_dispatcher.enable = True
        self._js[device_id].open()

    def close(self, device_id):
        """Close the device"""
        self._event_dispatcher.enable = False
        self._js[device_id].close()

    def read(self, device_id):
        """Read input from the selected device."""
        return self._js[device_id].read()

    def _dispatch_events(self, device_id, event):
        self._js[device_id].add_event(event)

    def devices(self):
        """List all the available devices."""
        logger.info("Looking for devices")
        names = []
        if len(self._devices) == 0:
            nbrOfInputs = sdl2.joystick.SDL_NumJoysticks()
            logger.info("Found {} devices".format(nbrOfInputs))
            for sdl_index in range(0, nbrOfInputs):
                j = sdl2.joystick.SDL_JoystickOpen(sdl_index)
                name = sdl2.joystick.SDL_JoystickName(j).decode("UTF-8")
                if names.count(name) > 0:
                    name = "{0} #{1}".format(name, names.count(name) + 1)
                sdl_id = sdl2.joystick.SDL_JoystickInstanceID(j)
                self._devices.append({"id": sdl_id, "name": name})
                self._js[sdl_id] = _JS(sdl_index, sdl_id, name)
                names.append(name)
                sdl2.joystick.SDL_JoystickClose(j)
        return self._devices



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2021 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301, USA.
"""
Sets up logging for the the full pose of the Crazyflie
"""
import logging
import math

from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.utils.callbacks import Caller

__author__ = 'Bitcraze AB'
__all__ = ['PoseLogger']

logger = logging.getLogger(__name__)


class PoseLogger:
    LOG_NAME_ESTIMATE_X = 'stateEstimate.x'
    LOG_NAME_ESTIMATE_Y = 'stateEstimate.y'
    LOG_NAME_ESTIMATE_Z = 'stateEstimate.z'
    LOG_NAME_ESTIMATE_ROLL = 'stateEstimate.roll'
    LOG_NAME_ESTIMATE_PITCH = 'stateEstimate.pitch'
    LOG_NAME_ESTIMATE_YAW = 'stateEstimate.yaw'
    NO_POSE = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)

    def __init__(self, cf: Crazyflie) -> None:
        self._cf = cf
        self._cf.connected.add_callback(self._connected)
        self._cf.disconnected.add_callback(self._disconnected)

        self.data_received_cb = Caller()
        self.error_cb = Caller()

        # The pose is an array containing
        # X, Y, Z, roll, pitch, yaw
        # roll, pitch and yaw in degrees
        self.pose = self.NO_POSE

    @property
    def position(self):
        """Get the position part of the full pose"""
        return self.pose[0:3]

    @property
    def rpy(self):
        """Get the roll, pitch and yaw of the full pose in degrees"""
        return self.pose[3:6]

    @property
    def rpy_rad(self):
        """Get the roll, pitch and yaw of the full pose in radians"""
        return [math.radians(self.pose[3]), math.radians(self.pose[4]), math.radians(self.pose[5])]

    def _connected(self, link_uri) -> None:
        logConf = LogConfig("Pose", 40)
        logConf.add_variable(self.LOG_NAME_ESTIMATE_X, "float")
        logConf.add_variable(self.LOG_NAME_ESTIMATE_Y, "float")
        logConf.add_variable(self.LOG_NAME_ESTIMATE_Z, "float")
        logConf.add_variable(self.LOG_NAME_ESTIMATE_ROLL, "float")
        logConf.add_variable(self.LOG_NAME_ESTIMATE_PITCH, "float")
        logConf.add_variable(self.LOG_NAME_ESTIMATE_YAW, "float")

        try:
            self._cf.log.add_config(logConf)
            logConf.data_received_cb.add_callback(self._data_received)
            logConf.error_cb.add_callback(self._error)
            logConf.start()
        except KeyError as e:
            logger.warning(str(e))
        except AttributeError as e:
            logger.warning(str(e))

    def _disconnected(self, link_uri) -> None:
        self.pose = self.NO_POSE

    def _data_received(self, timestamp, data, logconf) -> None:
        self.pose = (
            data[self.LOG_NAME_ESTIMATE_X],
            data[self.LOG_NAME_ESTIMATE_Y],
            data[self.LOG_NAME_ESTIMATE_Z],
            data[self.LOG_NAME_ESTIMATE_ROLL],
            data[self.LOG_NAME_ESTIMATE_PITCH],
            data[self.LOG_NAME_ESTIMATE_YAW],
        )

        self.data_received_cb.call(self, self.pose)

    def _error(self, log_conf, msg) -> None:
        self.error_cb.call(self, msg)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
Toolbox for showing packets that is sent via the communication link when
debugging.
"""
import os
from time import time
from binascii import hexlify

from PyQt6 import QtWidgets
from PyQt6 import uic
from PyQt6.QtCore import pyqtSignal
from PyQt6.QtCore import pyqtSlot
from PyQt6.QtCore import Qt

import cfclient
from cfclient.ui.tab_toolbox import TabToolbox

__author__ = 'Bitcraze AB'
__all__ = ['CrtpSharkToolbox']

param_tab_class = uic.loadUiType(cfclient.module_path + "/ui/tabs/crtpSharkToolbox.ui")[0]


class CrtpSharkToolbox(TabToolbox, param_tab_class):
    """Show packets that is sent vie the communication link"""
    nameModified = pyqtSignal()
    _incoming_packet_signal = pyqtSignal(object)
    _outgoing_packet_signal = pyqtSignal(object)

    def __init__(self, helper):
        super(CrtpSharkToolbox, self).__init__(helper, 'Crtp sniffer')
        self.setupUi(self)

        # Init the tree widget
        self.logTree.setHeaderLabels(['ms', 'Direction', 'Port/Chan', 'Data'])

        # Connect GUI signals
        self.clearButton.clicked.connect(self.clearLog)
        self.saveButton.clicked.connect(self._save_data)

        self._incoming_packet_signal.connect(lambda p: self._packet("IN", p))
        self._outgoing_packet_signal.connect(lambda p: self._packet("OUT", p))
        self._ms_offset = int(round(time() * 1000))

        self._data = []

    def _packet(self, dir, pk):
        if self.masterCheck.isChecked() and \
           not (pk.port == 15 and pk.channel == 3):
            line = QtWidgets.QTreeWidgetItem()

            ms_diff = int(round(time() * 1000)) - self._ms_offset
            line.setData(0, Qt.ItemDataRole.DisplayRole, "%d" % ms_diff)
            line.setData(1, Qt.ItemDataRole.DisplayRole, "%s" % dir)
            line.setData(2, Qt.ItemDataRole.DisplayRole, "%d/%d" % (pk.port, pk.channel))

            line.setData(3, Qt.ItemDataRole.DisplayRole, hexlify(pk.data).decode('utf8'))

            s = "%d, %s, %d/%d, %s" % (ms_diff, dir, pk.port, pk.channel,
                                       hexlify(pk.data).decode('utf8'))
            self._data.append(s)

            self.logTree.addTopLevelItem(line)
            self.logTree.scrollToItem(line)

    @pyqtSlot()
    def clearLog(self):
        self.logTree.clear()
        self._data = []

    def _incoming_packet(self, pk):
        self._incoming_packet_signal.emit(pk)

    def _outgoing_packet(self, pk):
        self._outgoing_packet_signal.emit(pk)

    def enable(self):
        self._helper.cf.packet_received.add_callback(self._incoming_packet)
        self._helper.cf.packet_sent.add_callback(self._outgoing_packet)

    def disable(self):
        self._helper.cf.packet_received.remove_callback(self._incoming_packet)
        self._helper.cf.packet_sent.remove_callback(self._outgoing_packet)

    def _save_data(self):
        dir = os.path.join(cfclient.config_path, "logdata")
        fname = os.path.join(dir, "shark_data.csv")
        if not os.path.exists(dir):
            os.makedirs(dir)
        f = open(fname, 'w')
        for s in self._data:
            f.write("%s\n" % s)
        f.close()



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2015 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.

"""
Server used to connect to a Crazyflie using ZMQ.
"""

import sys
import os
import logging
import signal
import zmq
import queue
from threading import Thread
import cflib.crtp
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig

import cfclient

if os.name == 'posix':
    print('Disabling standard output for libraries!')
    stdout = os.dup(1)
    os.dup2(os.open('/dev/null', os.O_WRONLY), 1)
    sys.stdout = os.fdopen(stdout, 'w')

# set SDL to use the dummy NULL video driver,
#   so it doesn't need a windowing system.
os.environ["SDL_VIDEODRIVER"] = "dummy"

# Main command socket for control (ping/pong)
ZMQ_SRV_PORT = 0
# Log data socket (publish)
ZMQ_LOG_PORT = 1
# Param value updated (publish)
ZMQ_PARAM_PORT = 2
# Async event for connection, like connection lost (publish)
ZMQ_CONN_PORT = 3
# Control set-poins for Crazyflie (pull)
ZMQ_CTRL_PORT = 4

# Timeout before giving up when verifying param write
PARAM_TIMEOUT = 2
# Timeout before giving up connection
CONNECT_TIMEOUT = 5
# Timeout before giving up adding/starting log config
LOG_TIMEOUT = 10

logger = logging.getLogger(__name__)


class _SrvThread(Thread):

    def __init__(self, socket, log_socket, param_socket, conn_socket, cf,
                 *args):
        super(_SrvThread, self).__init__(*args)
        self._socket = socket
        self._log_socket = log_socket
        self._param_socket = param_socket
        self._conn_socket = conn_socket
        self._cf = cf

        self._cf.connected.add_callback(self._connected)
        self._cf.connection_failed.add_callback(self._connection_failed)
        self._cf.connection_lost.add_callback(self._connection_lost)
        self._cf.disconnected.add_callback(self._disconnected)
        self._cf.connection_requested.add_callback(self._connection_requested)
        self._cf.param.all_updated.add_callback(self._tocs_updated)
        self._cf.param.all_update_callback.add_callback(self._all_param_update)

        self._conn_queue = queue.Queue(1)
        self._param_queue = queue.Queue(1)
        self._log_started_queue = queue.Queue(1)
        self._log_added_queue = queue.Queue(1)

        self._logging_configs = {}

    def _connection_requested(self, uri):
        conn_ev = {"version": 1, "event": "requested", "uri": uri}
        self._conn_socket.send_json(conn_ev)

    def _connected(self, uri):
        conn_ev = {"version": 1, "event": "connected", "uri": uri}
        self._conn_socket.send_json(conn_ev)

    def _connection_failed(self, uri, msg):
        logger.info("Connection failed to {}: {}".format(uri, msg))
        resp = {"version": 1, "status": 1, "msg": msg}
        self._conn_queue.put_nowait(resp)
        conn_ev = {"version": 1, "event": "failed", "uri": uri, "msg": msg}
        self._conn_socket.send_json(conn_ev)

    def _connection_lost(self, uri, msg):
        conn_ev = {"version": 1, "event": "lost", "uri": uri, "msg": msg}
        self._conn_socket.send_json(conn_ev)

    def _disconnected(self, uri):
        conn_ev = {"version": 1, "event": "disconnected", "uri": uri}
        self._conn_socket.send_json(conn_ev)

    def _tocs_updated(self):
        # First do the log
        log_toc = self._cf.log.toc.toc
        log = {}
        for group in log_toc:
            log[group] = {}
            for name in log_toc[group]:
                log[group][name] = {"type": log_toc[group][name].ctype}
        # The the params
        param_toc = self._cf.param.toc.toc
        param = {}
        for group in param_toc:
            param[group] = {}
            for name in param_toc[group]:
                param[group][name] = {
                    "type": param_toc[group][name].ctype,
                    "access": "RW" if param_toc[group][
                        name].access == 0 else "RO",
                    "value": self._cf.param.values[group][name]}

        resp = {"version": 1, "status": 0, "log": log, "param": param}
        self._conn_queue.put_nowait(resp)

    def _handle_scanning(self):
        resp = {"version": 1}
        interfaces = cflib.crtp.scan_interfaces()
        resp["interfaces"] = []
        for i in interfaces:
            resp["interfaces"].append({"uri": i[0], "info": i[1]})
        return resp

    def _handle_connect(self, uri):
        self._cf.open_link(uri)
        return self._conn_queue.get(block=True)

    def _logging_started(self, conf, started):
        out = {"version": 1, "name": conf.name}
        if started:
            out["event"] = "started"
        else:
            out["event"] = "stopped"
        self._log_socket.send_json(out)
        self._log_started_queue.put_nowait(started)

    def _logging_added(self, conf, added):
        out = {"version": 1, "name": conf.name}
        if added:
            out["event"] = "created"
        else:
            out["event"] = "deleted"
        self._log_socket.send_json(out)
        self._log_added_queue.put_nowait(added)

    def _handle_logging(self, data):
        resp = {"version": 1}
        if data["action"] == "create":
            lg = LogConfig(data["name"], data["period"])
            for v in data["variables"]:
                lg.add_variable(v)
            lg.started_cb.add_callback(self._logging_started)
            lg.added_cb.add_callback(self._logging_added)
            try:
                lg.data_received_cb.add_callback(self._logdata_callback)
                self._logging_configs[data["name"]] = lg
                self._cf.log.add_config(lg)
                lg.create()
                self._log_added_queue.get(block=True, timeout=LOG_TIMEOUT)
                resp["status"] = 0
            except KeyError as e:
                resp["status"] = 1
                resp["msg"] = str(e)
            except AttributeError as e:
                resp["status"] = 2
                resp["msg"] = str(e)
            except queue.Empty:
                resp["status"] = 3
                resp["msg"] = "Log configuration did not start"
        if data["action"] == "start":
            try:
                self._logging_configs[data["name"]].start()
                self._log_started_queue.get(block=True, timeout=LOG_TIMEOUT)
                resp["status"] = 0
            except KeyError as e:
                resp["status"] = 1
                resp["msg"] = "{} config not found".format(str(e))
            except queue.Empty:
                resp["status"] = 2
                resp["msg"] = "Log configuration did not stop"
        if data["action"] == "stop":
            try:
                self._logging_configs[data["name"]].stop()
                self._log_started_queue.get(block=True, timeout=LOG_TIMEOUT)
                resp["status"] = 0
            except KeyError as e:
                resp["status"] = 1
                resp["msg"] = "{} config not found".format(str(e))
            except queue.Empty:
                resp["status"] = 2
                resp["msg"] = "Log configuration did not stop"
        if data["action"] == "delete":
            try:
                self._logging_configs[data["name"]].delete()
                self._log_added_queue.get(block=True, timeout=LOG_TIMEOUT)
                resp["status"] = 0
            except KeyError as e:
                resp["status"] = 1
                resp["msg"] = "{} config not found".format(str(e))
            except queue.Empty:
                resp["status"] = 2
                resp["msg"] = "Log configuration did not stop"

        return resp

    def _handle_param(self, data):
        resp = {"version": 1}
        group = data["name"].split(".")[0]
        name = data["name"].split(".")[1]
        self._cf.param.add_update_callback(group=group, name=name,
                                           cb=self._param_callback)
        try:
            self._cf.param.set_value(data["name"], str(data["value"]))
            answer = self._param_queue.get(block=True, timeout=PARAM_TIMEOUT)
            resp["name"] = answer["name"]
            resp["value"] = answer["value"]
            resp["status"] = 0
        except KeyError as e:
            resp["status"] = 1
            resp["msg"] = str(e)
        except AttributeError as e:
            resp["status"] = 2
            resp["msg"] = str(e)
        except queue.Empty:
            resp["status"] = 3
            resp["msg"] = "Timeout when setting parameter" \
                          "{}".format(data["name"])
        return resp

    def _all_param_update(self, name, value):
        resp = {"version": 1, "name": name, "value": value}
        self._param_socket.send_json(resp)

    def _param_callback(self, name, value):
        group = name.split(".")[0]
        name_short = name.split(".")[1]
        self._cf.param.remove_update_callback(group=group, name=name_short)
        self._param_queue.put_nowait({"name": name, "value": value})

    def _logdata_callback(self, ts, data, conf):
        out = {"version": 1, "name": conf.name, "event": "data",
               "timestamp": ts, "variables": {}}
        for d in data:
            out["variables"][d] = data[d]
        self._log_socket.send_json(out)

    def run(self):
        logger.info("Starting server thread")
        while True:
            # Wait for the command
            cmd = self._socket.recv_json()
            response = {"version": 1}
            logger.info("Got command {}".format(cmd))
            if cmd["cmd"] == "scan":
                response = self._handle_scanning()
            elif cmd["cmd"] == "connect":
                response = self._handle_connect(cmd["uri"])
            elif cmd["cmd"] == "disconnect":
                self._cf.close_link()
                response["status"] = 0
            elif cmd["cmd"] == "log":
                response = self._handle_logging(cmd)
            elif cmd["cmd"] == "param":
                response = self._handle_param(cmd)
            else:
                response["status"] = 0xFF
                response["msg"] = "Unknown command {}".format(cmd["cmd"])
            self._socket.send_json(response)


class _CtrlThread(Thread):

    def __init__(self, socket, cf, *args):
        super(_CtrlThread, self).__init__(*args)
        self._socket = socket
        self._cf = cf

    def run(self):
        while True:
            cmd = self._socket.recv_json()
            self._cf.commander.send_setpoint(cmd["roll"], cmd["pitch"],
                                             cmd["yaw"], cmd["thrust"])


class ZMQServer():
    """Crazyflie ZMQ server"""

    def __init__(self, base_url, base_port):
        """Start threads and bind ports"""
        cflib.crtp.init_drivers()
        self._cf = Crazyflie(ro_cache=None,
                             rw_cache=cfclient.config_path + "/cache")

        signal.signal(signal.SIGINT, signal.SIG_DFL)

        self._base_url = base_url
        self._context = zmq.Context()

        cmd_srv = self._bind_zmq_socket(zmq.REP, "cmd",
                                        base_port + ZMQ_SRV_PORT)
        log_srv = self._bind_zmq_socket(zmq.PUB, "log",
                                        base_port + ZMQ_LOG_PORT)
        param_srv = self._bind_zmq_socket(zmq.PUB, "param",
                                          base_port + ZMQ_PARAM_PORT)
        ctrl_srv = self._bind_zmq_socket(zmq.PULL, "ctrl",
                                         base_port + ZMQ_CTRL_PORT)
        conn_srv = self._bind_zmq_socket(zmq.PUB, "conn",
                                         base_port + ZMQ_CONN_PORT)

        self._scan_thread = _SrvThread(cmd_srv, log_srv, param_srv, conn_srv,
                                       self._cf)
        self._scan_thread.start()

        self._ctrl_thread = _CtrlThread(ctrl_srv, self._cf)
        self._ctrl_thread.start()

    def _bind_zmq_socket(self, pattern, name, port):
        srv = self._context.socket(pattern)
        srv_addr = "{}:{}".format(self._base_url, port)
        srv.bind(srv_addr)
        logger.info("Biding ZMQ {} server"
                    "at {}".format(name, srv_addr))
        return srv


def main():
    """Main Crazyflie ZMQ application"""
    import argparse

    parser = argparse.ArgumentParser(prog="cfzmq")
    parser.add_argument("-u", "--url", action="store", dest="url", type=str,
                        default="tcp://127.0.0.1",
                        help="URL where ZMQ will accept connections")
    parser.add_argument("-d", "--debug", action="store_true", dest="debug",
                        help="Enable debug output")
    parser.add_argument("-p", "--port", action="store", dest="port", type=int,
                        default=2000,
                        help="Base port to used for ZMQ sockets")
    (args, _) = parser.parse_known_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    ZMQServer(args.url, args.port)

    # CRTL-C to exit


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2014 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301, USA.
"""
Mux for controlling roll/pitch from one device (slave/student) and the rest
from a second device (master/teacher) with the possibility to take over
roll/pitch as well.
"""

import logging

from . import InputMux

__author__ = 'Bitcraze AB'
__all__ = ['TakeOverSelectiveMux']

logger = logging.getLogger(__name__)


class TakeOverSelectiveMux(InputMux):

    def __init__(self, *args):
        super(TakeOverSelectiveMux, self).__init__(*args)
        self._master = "Teacher"
        self._slave = "Student"
        self.name = "Teacher (RP)"
        self._devs = {self._master: None, self._slave: None}

        self._muxing = {
            self._master: ("thrust", "yaw", "estop", "alt1", "alt2",
                           "assistedControl", "exit"),
            self._slave: ("roll", "pitch")
        }

    def read(self):
        try:
            if self._devs[self._master] and self._devs[self._slave]:
                dm = self._devs[self._master].read()
                ds = self._devs[self._slave].read()
                if not dm.muxswitch:
                    for key in self._muxing[self._slave]:
                        dm.set(key, ds.get(key))

                return dm
            else:
                return None

        except Exception as e:
            logger.warning(e)
            return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2025 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#  02110-1301, USA.

"""
Basic tab to be able to set (and test) colors in the LED-ring.
"""

import logging

from PyQt6 import QtGui, uic
from PyQt6.QtCore import pyqtSignal
from PyQt6 import QtWidgets

import cfclient
from cfclient.ui.tab_toolbox import TabToolbox

from cflib.crazyflie.mem import MemoryElement

__author__ = 'Bitcraze AB'
__all__ = ['LEDRingTab']

logger = logging.getLogger(__name__)

led_ring_tab_class = uic.loadUiType(cfclient.module_path + "/ui/tabs/ledRingTab.ui")[0]


class LEDRingTab(TabToolbox, led_ring_tab_class):
    """Tab for plotting logging data"""

    _connected_signal = pyqtSignal(str)
    _disconnected_signal = pyqtSignal(str)

    def __init__(self, helper):
        super(LEDRingTab, self).__init__(helper, 'LED Ring')
        self.setupUi(self)

        # LED-ring effect dropdown and headlight checkbox
        self._ledring_nbr_effects = 0

        # Connect the headlight checkbox
        self._led_ring_headlight.clicked.connect(
            lambda enabled: self._helper.cf.param.set_value("ring.headlightEnable", int(enabled)))

        # Update headlight when param changes
        self._helper.cf.param.add_update_callback(
            group="ring", name="headlightEnable",
            cb=lambda name, checked: self._led_ring_headlight.setChecked(bool(int(checked))))

        # Update LED-ring effect when param changes
        self._helper.cf.param.add_update_callback(
            group="ring", name="effect",
            cb=self._ring_effect_updated)

        # Populate dropdown when all params are updated
        self._helper.cf.param.all_updated.add_callback(self._ring_populate_dropdown)

        # Always wrap callbacks from Crazyflie API though QT Signal/Slots
        # to avoid manipulating the UI when rendering it
        self._connected_signal.connect(self._connected)
        self._disconnected_signal.connect(self._disconnected)

        # Connect the Crazyflie API callbacks to the signals
        self._helper.cf.connected.add_callback(
            self._connected_signal.emit)

        self._helper.cf.disconnected.add_callback(
            self._disconnected_signal.emit)

        self._btns = [self._u1,
                      self._u2,
                      self._u3,
                      self._u4,
                      self._u5,
                      self._u6,
                      self._u7,
                      self._u8,
                      self._u9,
                      self._u10,
                      self._u11,
                      self._u12]

        self._intensity = self._intensity_slider.value()

        self._u1.clicked.connect(lambda: self._select(0))
        self._u2.clicked.connect(lambda: self._select(1))
        self._u3.clicked.connect(lambda: self._select(2))
        self._u4.clicked.connect(lambda: self._select(3))
        self._u5.clicked.connect(lambda: self._select(4))
        self._u6.clicked.connect(lambda: self._select(5))
        self._u7.clicked.connect(lambda: self._select(6))
        self._u8.clicked.connect(lambda: self._select(7))
        self._u9.clicked.connect(lambda: self._select(8))
        self._u10.clicked.connect(lambda: self._select(9))
        self._u11.clicked.connect(lambda: self._select(10))
        self._u12.clicked.connect(lambda: self._select(11))

        self._mem = None

        self._intensity_slider.valueChanged.connect(self._intensity_change)
        self._intensity_slider.valueChanged.connect(
            self._intensity_spin.setValue)
        self._intensity_spin.valueChanged.connect(
            self._intensity_slider.setValue)

        self._helper.inputDeviceReader.alt1_updated.add_callback(self.alt1_updated)
        self._helper.inputDeviceReader.alt2_updated.add_callback(self.alt2_updated)

        self._led_ring_effect.setEnabled(False)
        self._led_ring_headlight.setEnabled(False)

    def _select(self, nbr):
        col = QtGui.QColor()

        if self._mem:
            led = self._mem.leds[nbr]
            col = QtGui.QColor.fromRgb(led.r, led.g, led.b)

        col = QtWidgets.QColorDialog.getColor(col)

        if col.isValid() and self._mem:
            r, g, b = col.red(), col.green(), col.blue()
            self._mem.leds[nbr].set(r=r, g=g, b=b)

            brightness = 0.299 * r + 0.587 * g + 0.114 * b
            text_color = "white" if brightness < 128 else "black"

            self.sender().setStyleSheet(
                f"background-color: rgb({r}, {g}, {b}); color: {text_color};"
            )

            self._write_led_output()

    def _intensity_change(self, value):
        self._intensity = value
        self._write_led_output()

    def _write_led_output(self):
        if self._mem:
            for led in self._mem.leds:
                led.intensity = self._intensity
            self._mem.write_data(self._led_write_done)
        else:
            logger.info("No LED-ring memory found!")

    def _led_write_done(self, mem, addr):
        logger.info("LED write done callback")

    def _connected(self, link_uri):
        """Callback when the Crazyflie has been connected"""
        mems = self._helper.cf.mem.get_mems(MemoryElement.TYPE_DRIVER_LED)
        if len(mems) > 0:
            self._mem = mems[0]
            logger.info(self._mem)

        if self._mem:
            for btn in self._btns:
                btn.setEnabled(True)
                btn.setStyleSheet("background-color: black; color: white")
                self._intensity_slider.setEnabled(True)
                self._intensity_spin.setEnabled(True)

        self._led_ring_effect.setEnabled(True)
        self._led_ring_headlight.setEnabled(True)

    def _disconnected(self, link_uri):
        """Callback for when the Crazyflie has been disconnected"""
        for btn in self._btns:
            btn.setEnabled(False)
            btn.setStyleSheet("background-color: none")
            self._intensity_slider.setEnabled(False)
            self._intensity_spin.setEnabled(False)
            self._intensity_slider.setValue(100)

        self._led_ring_effect.setEnabled(False)
        self._led_ring_headlight.setEnabled(False)

    def _ring_populate_dropdown(self):
        try:
            nbr = int(self._helper.cf.param.values["ring"]["neffect"])
            current = int(self._helper.cf.param.values["ring"]["effect"])
        except KeyError:
            return

        self._ring_effect = current
        self._ledring_nbr_effects = nbr

        hardcoded_names = {
            0: "Off",
            1: "White spinner",
            2: "Color spinner",
            3: "Tilt effect",
            4: "Brightness effect",
            5: "Color spinner 2",
            6: "Double spinner",
            7: "Solid color effect",
            8: "Factory test",
            9: "Battery status",
            10: "Boat lights",
            11: "Alert",
            12: "Gravity",
            13: "LED tab",
            14: "Color fader",
            15: "Link quality",
            16: "Location server status",
            17: "Sequencer",
            18: "Lighthouse quality",
        }

        self._led_ring_effect.clear()
        for i in range(nbr + 1):
            name = "{}: ".format(i)
            name += hardcoded_names.get(i, "N/A")
            self._led_ring_effect.addItem(name, i)

        self._led_ring_effect.currentIndexChanged.connect(self._ring_effect_changed)
        self._led_ring_effect.setCurrentIndex(current)
        self._led_ring_effect.setEnabled(int(self._helper.cf.param.values["deck"]["bcLedRing"]) == 1)
        self._led_ring_headlight.setEnabled(int(self._helper.cf.param.values["deck"]["bcLedRing"]) == 1)

        try:
            self._helper.cf.param.set_value("ring.effect", "13")
            self._led_ring_effect.setCurrentIndex(13)
            self._ring_effect = 13
            logger.info("Initialized LED ring to 'LED tab' mode (effect 13).")
        except Exception as e:
            logger.warning(f"Could not set LED tab effect on connect: {e}")

    def _ring_effect_changed(self, index):
        self._ring_effect = index
        if index > -1:
            i = self._led_ring_effect.itemData(index)
            if i != int(self._helper.cf.param.values["ring"]["effect"]):
                self._helper.cf.param.set_value("ring.effect", str(i))

            if i == 13:
                self._intensity_slider.setEnabled(True)
                self._intensity_spin.setEnabled(True)
            else:
                self._intensity_slider.setEnabled(False)
                self._intensity_spin.setEnabled(False)

    def _ring_effect_updated(self, name, value):
        if self._helper.cf.param.is_updated:
            self._led_ring_effect.setCurrentIndex(int(value))

    def alt1_updated(self, state):
        if state:
            new_index = (self._ring_effect+1) % (self._ledring_nbr_effects+1)
            self._helper.cf.param.set_value("ring.effect", str(new_index))

    def alt2_updated(self, state):
        self._helper.cf.param.set_value("ring.headlightEnable", str(state))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc., 51
#  Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
Module to read input devices and send controls to the Crazyflie.

This module reads input from joysticks or other input devices and sends control
set-points to the Crazyflie. It can be configured in the UI.

Various drivers can be used to read input device data. Currently is uses the
PySDL2 driver, but in the future native support will be provided for Linux and
Windows drivers.

The input device's axes and buttons are mapped to software inputs using a
configuration file.
"""
import os
import re
import glob
import traceback
import logging
import shutil

from . import inputreaders as readers
from . import inputinterfaces as interfaces

import cfclient
from cfclient.utils.config import Config
from cfclient.utils.config_manager import ConfigManager

from cfclient.utils.periodictimer import PeriodicTimer
from cflib.utils.callbacks import Caller
from .mux.nomux import NoMux
from .mux.takeovermux import TakeOverMux
from .mux.takeoverselectivemux import TakeOverSelectiveMux

__author__ = 'Bitcraze AB'
__all__ = ['JoystickReader']

logger = logging.getLogger(__name__)

MAX_THRUST = 65000
INITAL_TAGET_HEIGHT = 0.4
MAX_TARGET_HEIGHT = 1.0
MIN_TARGET_HEIGHT = 0.03
MIN_HOVER_HEIGHT = 0.20
INPUT_READ_PERIOD = 0.01


class JoystickReader(object):
    """
    Thread that will read input from devices/joysticks and send control-set
    points to the Crazyflie
    """
    inputConfig = []

    ASSISTED_CONTROL_ALTHOLD = 0
    ASSISTED_CONTROL_POSHOLD = 1
    ASSISTED_CONTROL_HEIGHTHOLD = 2
    ASSISTED_CONTROL_HOVER = 3

    def __init__(self, do_device_discovery=True):
        self._input_device = None

        self._mux = [NoMux(self), TakeOverSelectiveMux(self),
                     TakeOverMux(self)]
        # Set NoMux as default
        self._selected_mux = self._mux[0]

        self.min_thrust = 0
        self.max_thrust = 0
        self._thrust_slew_rate = 0
        self.thrust_slew_enabled = False
        self.thrust_slew_limit = 0
        self.has_pressure_sensor = False
        self._hover_max_height = MAX_TARGET_HEIGHT

        self.max_rp_angle = 0
        self.max_yaw_rate = 0
        try:
            self.set_assisted_control(Config().get("assistedControl"))
        except KeyError:
            self.set_assisted_control(JoystickReader.ASSISTED_CONTROL_ALTHOLD)

        self._old_thrust = 0
        self._old_raw_thrust = 0
        self.springy_throttle = True
        self._target_height = INITAL_TAGET_HEIGHT

        self.trim_roll = Config().get("trim_roll")
        self.trim_pitch = Config().get("trim_pitch")
        self._rp_dead_band = 0.1

        self._input_map = None

        if Config().get("flightmode") == "Normal":
            self.max_yaw_rate = Config().get("normal_max_yaw")
            self.max_rp_angle = Config().get("normal_max_rp")
            # Values are stored at %, so use the functions to set the values
            self.min_thrust = Config().get("normal_min_thrust")
            self.max_thrust = Config().get("normal_max_thrust")
            self.thrust_slew_limit = Config().get("normal_slew_limit")
            self.thrust_slew_rate = Config().get("normal_slew_rate")
        else:
            self.max_yaw_rate = Config().get("max_yaw")
            self.max_rp_angle = Config().get("max_rp")
            # Values are stored at %, so use the functions to set the values
            self.min_thrust = Config().get("min_thrust")
            self.max_thrust = Config().get("max_thrust")
            self.thrust_slew_limit = Config().get("slew_limit")
            self.thrust_slew_rate = Config().get("slew_rate")

        self._dev_blacklist = None
        if len(Config().get("input_device_blacklist")) > 0:
            self._dev_blacklist = re.compile(
                Config().get("input_device_blacklist"))
        logger.info("Using device blacklist [{}]".format(
            Config().get("input_device_blacklist")))

        self._available_devices = {}

        # TODO: The polling interval should be set from config file
        self._read_timer = PeriodicTimer(INPUT_READ_PERIOD, self.read_input)

        if do_device_discovery:
            self._discovery_timer = PeriodicTimer(1.0,
                                                  self._do_device_discovery)
            self._discovery_timer.start()

        # Check if user config exists, otherwise copy files
        if not os.path.exists(ConfigManager().configs_dir):
            logger.info("No user config found, copying dist files")
            os.makedirs(ConfigManager().configs_dir)

        for f in glob.glob(
                cfclient.module_path + "/configs/input/[A-Za-z]*.json"):
            dest = os.path.join(ConfigManager().
                                configs_dir, os.path.basename(f))
            if not os.path.isfile(dest):
                logger.debug("Copying %s", f)
                shutil.copy2(f, ConfigManager().configs_dir)

        ConfigManager().get_list_of_configs()

        self.input_updated = Caller()
        self.assisted_input_updated = Caller()
        self.heighthold_input_updated = Caller()
        self.hover_input_updated = Caller()
        self.rp_trim_updated = Caller()
        self.emergency_stop_updated = Caller()
        self.arm_updated = Caller()
        self.device_discovery = Caller()
        self.device_error = Caller()
        self.assisted_control_updated = Caller()
        self.alt1_updated = Caller()
        self.alt2_updated = Caller()

        # Call with 3 bools (rp_limiting, yaw_limiting, thrust_limiting)
        self.limiting_updated = Caller()

    def _get_device_from_name(self, device_name):
        """Get the raw device from a name"""
        for d in readers.devices():
            if d.name == device_name:
                return d
        return None

    def set_hover_max_height(self, height):
        self._hover_max_height = height

    def set_alt_hold_available(self, available):
        """Set if altitude hold is available or not (depending on HW)"""
        self.has_pressure_sensor = available

    def _do_device_discovery(self):
        devs = self.available_devices()

        # This is done so that devs can easily get access
        # to limits without creating lots of extra code
        for d in devs:
            d.input = self

        if len(devs):
            self.device_discovery.call(devs)
            self._discovery_timer.stop()

    def available_mux(self):
        return self._mux

    def set_mux(self, name=None, mux=None):
        old_mux = self._selected_mux
        if name:
            for m in self._mux:
                if m.name == name:
                    self._selected_mux = m
        elif mux:
            self._selected_mux = mux

        old_mux.close()

        logger.info("Selected MUX: {}".format(self._selected_mux.name))

    def set_assisted_control(self, mode):
        self._assisted_control = mode

    def get_assisted_control(self):
        return self._assisted_control

    def available_devices(self):
        """List all available and approved input devices.
        This function will filter available devices by using the
        blacklist configuration and only return approved devices."""
        devs = readers.devices()
        devs += interfaces.devices()
        approved_devs = []

        for dev in devs:
            if ((not self._dev_blacklist) or
                    (self._dev_blacklist and
                     not self._dev_blacklist.match(dev.name))):
                dev.input = self
                approved_devs.append(dev)

        return approved_devs

    def enableRawReading(self, device_name):
        """
        Enable raw reading of the input device with id deviceId. This is used
        to get raw values for setting up of input devices. Values are read
        without using a mapping.
        """
        if self._input_device:
            self._input_device.close()
            self._input_device = None

        for d in readers.devices():
            if d.name == device_name:
                self._input_device = d

        # Set the mapping to None to get raw values
        self._input_device.input_map = None
        self._input_device.open()

    def get_saved_device_mapping(self, device_name):
        """Return the saved mapping for a given device"""
        config = None
        device_config_mapping = Config().get("device_config_mapping")
        if device_name in list(device_config_mapping.keys()):
            config = device_config_mapping[device_name]

        logging.debug("For [{}] we recommend [{}]".format(device_name, config))
        return config

    def stop_raw_reading(self):
        """Disable raw reading of input device."""
        if self._input_device:
            self._input_device.close()
            self._input_device = None

    def read_raw_values(self):
        """ Read raw values from the input device."""
        [axes, buttons, mapped_values] = self._input_device.read(
            include_raw=True)
        dict_axes = {}
        dict_buttons = {}

        for i, a in enumerate(axes):
            dict_axes[i] = a

        for i, b in enumerate(buttons):
            dict_buttons[i] = b

        return [dict_axes, dict_buttons, mapped_values]

    def set_raw_input_map(self, input_map):
        """Set an input device map"""
        # TODO: Will not work!
        if self._input_device:
            self._input_device.input_map = input_map

    def set_input_map(self, device_name, input_map_name):
        """Load and set an input device map with the given name"""
        dev = self._get_device_from_name(device_name)
        settings = ConfigManager().get_settings(input_map_name)

        if settings:
            self.springy_throttle = settings["springythrottle"]
            self._rp_dead_band = settings["rp_dead_band"]
            self._input_map = ConfigManager().get_config(input_map_name)
        dev.input_map = self._input_map
        dev.input_map_name = input_map_name
        Config().get("device_config_mapping")[device_name] = input_map_name
        dev.set_dead_band(self._rp_dead_band)

    def start_input(self, device_name, role="Device", config_name=None):
        """
        Start reading input from the device with name device_name using config
        config_name. Returns True if device supports mapping, otherwise False
        """
        try:
            # device_id = self._available_devices[device_name]
            # Check if we supplied a new map, if not use the preferred one
            device = self._get_device_from_name(device_name)
            self._selected_mux.add_device(device, role)
            # Update the UI with the limiting for this device
            self.limiting_updated.call(device.limit_rp,
                                       device.limit_yaw,
                                       device.limit_thrust)
            self._read_timer.start()
            return device.supports_mapping
        except Exception:
            self.device_error.call(
                "Error while opening/initializing  input device\n\n%s" %
                (traceback.format_exc()))

        if not self._input_device:
            self.device_error.call(
                "Could not find device {}".format(device_name))
        return False

    def resume_input(self):
        self._selected_mux.resume()
        self._read_timer.start()

    def pause_input(self, device_name=None):
        """Stop reading from the input device."""
        self._read_timer.stop()
        self._selected_mux.pause()

    def _set_thrust_slew_rate(self, rate):
        self._thrust_slew_rate = rate
        if rate > 0:
            self.thrust_slew_enabled = True
        else:
            self.thrust_slew_enabled = False

    def _get_thrust_slew_rate(self):
        return self._thrust_slew_rate

    def read_input(self):
        """Read input data from the selected device"""
        try:
            data = self._selected_mux.read()

            if data:
                if data.toggled.assistedControl:
                    if self._assisted_control == \
                            JoystickReader.ASSISTED_CONTROL_POSHOLD or \
                            self._assisted_control == \
                            JoystickReader.ASSISTED_CONTROL_HOVER:
                        if data.assistedControl and self._assisted_control != \
                                JoystickReader.ASSISTED_CONTROL_HOVER:
                            for d in self._selected_mux.devices():
                                d.limit_thrust = False
                                d.limit_rp = False
                        elif data.assistedControl:
                            for d in self._selected_mux.devices():
                                d.limit_thrust = True
                                d.limit_rp = False
                        else:
                            for d in self._selected_mux.devices():
                                d.limit_thrust = True
                                d.limit_rp = True
                    if self._assisted_control == \
                            JoystickReader.ASSISTED_CONTROL_ALTHOLD:
                        self.assisted_control_updated.call(
                                            data.assistedControl)
                    if ((self._assisted_control ==
                            JoystickReader.ASSISTED_CONTROL_HEIGHTHOLD) or
                            (self._assisted_control ==
                             JoystickReader.ASSISTED_CONTROL_HOVER)):
                        try:
                            self.assisted_control_updated.call(
                                                data.assistedControl)
                            if not data.assistedControl:
                                # Reset height controller state to initial
                                # target height both in the UI and in the
                                # Crazyflie.
                                # TODO: Implement a proper state update of the
                                #       input layer
                                self.heighthold_input_updated.\
                                    call(0, 0,
                                         0, INITAL_TAGET_HEIGHT)
                                self.hover_input_updated.\
                                    call(0, 0,
                                         0, INITAL_TAGET_HEIGHT)
                        except Exception as e:
                            logger.warning(
                                "Exception while doing callback from "
                                "input-device for assited "
                                "control: {}".format(e))

                if data.toggled.estop:
                    try:
                        self.emergency_stop_updated.call(data.estop)
                    except Exception as e:
                        logger.warning("Exception while doing callback from"
                                       "input-device for estop: {}".format(e))
                if data.toggled.arm and data._prev_btn_values["arm"]:
                    try:
                        self.arm_updated.call(data.arm)
                    except Exception as e:
                        logger.warning("Exception while doing callback from"
                                       "input-device for arm: {}".format(e))
                if data.toggled.alt1:
                    try:
                        self.alt1_updated.call(data.alt1)
                    except Exception as e:
                        logger.warning("Exception while doing callback from"
                                       "input-device for alt1: {}".format(e))
                if data.toggled.alt2:
                    try:
                        self.alt2_updated.call(data.alt2)
                    except Exception as e:
                        logger.warning("Exception while doing callback from"
                                       "input-device for alt2: {}".format(e))

                # Reset height target when height-hold is not selected
                if not data.assistedControl or \
                        (self._assisted_control !=
                         JoystickReader.ASSISTED_CONTROL_HEIGHTHOLD and
                         self._assisted_control !=
                         JoystickReader.ASSISTED_CONTROL_HOVER):
                    self._target_height = INITAL_TAGET_HEIGHT

                if self._assisted_control == \
                        JoystickReader.ASSISTED_CONTROL_POSHOLD \
                        and data.assistedControl:
                    vx = data.roll
                    vy = data.pitch
                    vz = data.thrust
                    yawrate = -data.yaw
                    # The odd use of vx and vy is to map forward on the
                    # physical joystick to positive X-axis
                    self.assisted_input_updated.call(vy, -vx, vz, yawrate)
                elif self._assisted_control == \
                        JoystickReader.ASSISTED_CONTROL_HOVER \
                        and data.assistedControl:
                    vx = data.roll
                    vy = data.pitch

                    # Scale thrust to a value between -1.0 to 1.0
                    vz = (data.thrust - 32767) / 32767.0
                    # Integrate velocity setpoint
                    self._target_height += vz * INPUT_READ_PERIOD
                    # Cap target height
                    if self._target_height > self._hover_max_height:
                        self._target_height = self._hover_max_height
                    if self._target_height < MIN_HOVER_HEIGHT:
                        self._target_height = MIN_HOVER_HEIGHT

                    yawrate = -data.yaw
                    # The odd use of vx and vy is to map forward on the
                    # physical joystick to positive X-axis
                    self.hover_input_updated.call(vy, -vx, yawrate,
                                                  self._target_height)
                else:
                    # Update the user roll/pitch trim from device
                    if data.toggled.pitchNeg and data.pitchNeg:
                        self.trim_pitch -= .2
                    if data.toggled.pitchPos and data.pitchPos:
                        self.trim_pitch += .2
                    if data.toggled.rollNeg and data.rollNeg:
                        self.trim_roll -= .2
                    if data.toggled.rollPos and data.rollPos:
                        self.trim_roll += .2

                    if data.toggled.pitchNeg or data.toggled.pitchPos or \
                            data.toggled.rollNeg or data.toggled.rollPos:
                        self.rp_trim_updated.call(self.trim_roll,
                                                  self.trim_pitch)

                    if self._assisted_control == \
                            JoystickReader.ASSISTED_CONTROL_HEIGHTHOLD \
                            and data.assistedControl:
                        roll = data.roll + self.trim_roll
                        pitch = data.pitch + self.trim_pitch
                        yawrate = -data.yaw
                        # Scale thrust to a value between -1.0 to 1.0
                        vz = (data.thrust - 32767) / 32767.0
                        # Integrate velocity setpoint
                        self._target_height += vz * INPUT_READ_PERIOD
                        # Cap target height
                        if self._target_height > self._hover_max_height:
                            self._target_height = self._hover_max_height
                        if self._target_height < MIN_TARGET_HEIGHT:
                            self._target_height = MIN_TARGET_HEIGHT
                        self.heighthold_input_updated.call(roll, -pitch,
                                                           yawrate,
                                                           self._target_height)
                    else:
                        # Using alt hold the data is not in a percentage
                        if not data.assistedControl:
                            data.thrust = JoystickReader.p2t(data.thrust)

                        # Thrust might be <0 here, make sure it's not otherwise
                        # we'll get an error.
                        if data.thrust < 0:
                            data.thrust = 0
                        if data.thrust > 0xFFFF:
                            data.thrust = 0xFFFF

                        self.input_updated.call(data.roll + self.trim_roll,
                                                data.pitch + self.trim_pitch,
                                                data.yaw, data.thrust)
            else:
                self.input_updated.call(0, 0, 0, 0)
        except Exception:
            logger.warning("Exception while reading inputdevice: %s",
                           traceback.format_exc())
            self.device_error.call("Error reading from input device\n\n%s" %
                                   traceback.format_exc())
            self.input_updated.call(0, 0, 0, 0)
            self._read_timer.stop()

    @staticmethod
    def p2t(percentage):
        """Convert a percentage to raw thrust"""
        return int(MAX_THRUST * (percentage / 100.0))

    thrust_slew_rate = property(_get_thrust_slew_rate, _set_thrust_slew_rate)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2013 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.

#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.

"""
Used to write log data to files.
"""

import os
import datetime

import logging

import cfclient

__author__ = 'Bitcraze AB'
__all__ = ['LogWriter']

logger = logging.getLogger(__name__)


class LogWriter():
    """Create a writer for a specific log block"""

    def __init__(self, logblock, connected_ts=None, directory=None):
        """Initialize the writer"""
        self._block = logblock
        self._dir = directory
        self._connected_ts = connected_ts

        self._dir = os.path.join(cfclient.config_path, "logdata",
                                 connected_ts.strftime("%Y%m%dT%H-%M-%S"))
        self._file = None
        self._header_written = False
        self._header_values = []
        self._filename = None

    def _write_header(self):
        """Write the header to the file"""
        if not self._header_written:
            s = "Timestamp"
            for v in self._block.variables:
                s += "," + v.name
                self._header_values.append(v.name)
            s += '\n'
            self._file.write(s)
            self._header_written = True

    def _new_data(self, timestamp, data, logconf):
        """Callback when new data arrives from the Crazyflie"""
        if self._file:
            s = "%d" % timestamp
            for col in self._header_values:
                s += "," + str(data[col])
            s += '\n'
            self._file.write(s)

    def writing(self):
        """Return True if the file is open and we are using it,
        otherwise false"""
        return True if self._file else False

    def stop(self):
        """Stop the logging to file"""
        if self._file:
            self._file.close()
            self._file = None
            self._block.data_received_cb.remove_callback(self._new_data)
            logger.info("Stopped logging of block [%s] to file [%s]",
                        self._block.name, self._filename)
            self._header_values = []
            self._header_written = False

    def start(self):
        """Start the logging to file"""

        # Due to concurrency let's not check first, just create
        try:
            os.makedirs(self._dir)
        except OSError:
            logger.debug("logdata directory already exists")

        if not self._file:
            time_now = datetime.datetime.now()
            block_name_corr = self._block.name.replace('/', '-')
            name = "{0}-{1}.csv".format(block_name_corr,
                                        time_now.strftime(
                                            "%Y%m%dT%H-%M-%S"))
            self._filename = os.path.join(self._dir, name)
            self._file = open(self._filename, 'w')
            self._write_header()
            self._block.data_received_cb.add_callback(self._new_data)
            logger.info("Started logging of block [%s] to file [%s]",
                        self._block.name, self._filename)



# -*- coding: utf-8 -*-
#
# ,---------,       ____  _ __
# |  ,-^-,  |      / __ )(_) /_______________ _____  ___
# | (  O  ) |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
# | / ,--'  |    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#    +------`   /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
# Copyright (C) 2022-2023 Bitcraze AB
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, in version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Tab for tuning PID controller, mainly for larger quads.
"""

import logging

from PyQt6 import QtWidgets
from PyQt6 import uic
from PyQt6.QtCore import pyqtSignal, Qt
import time

import cfclient
from cfclient.ui.tab_toolbox import TabToolbox
from cfclient.ui.widgets.super_slider import SuperSlider
from cflib.crazyflie import Crazyflie, Param
from cflib.utils.callbacks import Syncer

__author__ = 'Bitcraze AB'
__all__ = ['TuningTab']

logger = logging.getLogger(__name__)

tuning_tab_class = uic.loadUiType(cfclient.module_path + "/ui/tabs/tuningTab.ui")[0]


class SliderParamMapper:
    def __init__(self, slider: SuperSlider, group: str, name: str):
        self.param_group = group
        self.param_name = name
        self.full_param_name = f'{group}.{name}'

        self.slider = slider
        self.slider.value_changed_cb.add_callback(self.slider_changed)

        # Prevents param changes that comes back from the CF to set the parameter and create a feedback loop
        self.receive_block_time = 0.0

        self.linked_mapper = None
        self.linked_checkbox: QtWidgets.QCheckBox = None

        self.cf = None

    def connected(self, cf: Crazyflie):
        self.cf = cf
        return self.param_group, self.param_name

    def disconnected(self):
        self.cf = None

    def link_with(self, other_mapper: 'SliderParamMapper', checkbox: QtWidgets.QCheckBox):
        self.linked_mapper = other_mapper
        self.linked_checkbox = checkbox

        other_mapper.linked_mapper = self
        other_mapper.linked_checkbox = checkbox

    def enable_ui(self, enabled):
        self.slider.setEnabled(enabled)

    # Called when the user has modified the value in the UI
    def slider_changed(self, value):
        if self.cf is not None:
            if self.cf.is_connected():
                self.receive_block_time = time.time() + 0.4
                self.cf.param.set_value(self.full_param_name, value)

        if self.linked_mapper is not None:
            self.linked_mapper.linked_update(value)

    # Called when a parameter in the CF has changed, this is also true if we initiated the param update
    def param_updated_cb(self, full_param_name, value):
        if time.time() > self.receive_block_time:
            self.slider.set_value(float(value))

            if self.linked_mapper is not None:
                self.linked_mapper.linked_update(value)

    # Called by another SliderParamMapper that is linked
    def linked_update(self, value):
        if self.linked_checkbox is not None and self.linked_checkbox.checkState() == Qt.CheckState.Checked:
            self.slider.set_value(float(value))
            if self.cf is not None and self.cf.is_connected():
                self.receive_block_time = time.time() + 0.4
                self.cf.param.set_value(self.full_param_name, value)


class TuningTab(TabToolbox, tuning_tab_class):
    """Tab for plotting logging data"""

    _connected_signal = pyqtSignal(str)
    _disconnected_signal = pyqtSignal(str)

    _param_updated_signal = pyqtSignal(str, object)

    def __init__(self, helper):
        super(TuningTab, self).__init__(helper, 'Tuning')
        self.setupUi(self)

        # Always wrap callbacks from Crazyflie API though QT Signal/Slots
        # to avoid manipulating the UI when rendering it
        self._connected_signal.connect(self._connected)
        self._disconnected_signal.connect(self._disconnected)
        self._param_updated_signal.connect(self._param_updated_cb)

        # Connect the Crazyflie API callbacks to the signals
        self._helper.cf.connected.add_callback(
            self._connected_signal.emit)

        self._helper.cf.disconnected.add_callback(
            self._disconnected_signal.emit)

        # A dictionary mapping the full param name to a SliderParamMapper
        self.mappers = self._set_up_param_mappings()

        self.store_button.clicked.connect(self._store_button_clicked)
        self.clear_stored_button.clicked.connect(self._clear_stored_button_clicked)
        self.default_values_button.clicked.connect(self._default_values_button_clicked)
        self._enable_ui_objects(False)

    def _set_up_param_mappings(self):
        mappers: dict[str, SliderParamMapper] = {}

        # Rate PID
        mappers.update(self._create_slider(self.grid_rate, 1, 1, 0, 1000, 'pid_rate', 'roll_kp'))
        mappers.update(self._create_slider(self.grid_rate, 1, 2, 0, 1000, 'pid_rate', 'roll_ki'))
        mappers.update(self._create_slider(self.grid_rate, 1, 3, 0, 10, 'pid_rate', 'roll_kd'))

        mappers.update(self._create_slider(self.grid_rate, 2, 1, 0, 1000, 'pid_rate', 'pitch_kp'))
        mappers.update(self._create_slider(self.grid_rate, 2, 2, 0, 1000, 'pid_rate', 'pitch_ki'))
        mappers.update(self._create_slider(self.grid_rate, 2, 3, 0, 10, 'pid_rate', 'pitch_kd'))

        mappers.update(self._create_slider(self.grid_rate, 3, 1, 0, 200, 'pid_rate', 'yaw_kp'))
        mappers.update(self._create_slider(self.grid_rate, 3, 2, 0, 100, 'pid_rate', 'yaw_ki'))
        mappers.update(self._create_slider(self.grid_rate, 3, 3, 0, 10, 'pid_rate', 'yaw_kd'))

        self._link(mappers, 'pid_rate.roll_kp', 'pid_rate.pitch_kp', self.rate_link_checkbox)
        self._link(mappers, 'pid_rate.roll_ki', 'pid_rate.pitch_ki', self.rate_link_checkbox)
        self._link(mappers, 'pid_rate.roll_kd', 'pid_rate.pitch_kd', self.rate_link_checkbox)

        # Attitude PID
        mappers.update(self._create_slider(self.grid_attitude, 1, 1, 0, 10, 'pid_attitude', 'roll_kp'))
        mappers.update(self._create_slider(self.grid_attitude, 1, 2, 0, 10, 'pid_attitude', 'roll_ki'))
        mappers.update(self._create_slider(self.grid_attitude, 1, 3, 0, 10, 'pid_attitude', 'roll_kd'))

        mappers.update(self._create_slider(self.grid_attitude, 2, 1, 0, 10, 'pid_attitude', 'pitch_kp'))
        mappers.update(self._create_slider(self.grid_attitude, 2, 2, 0, 10, 'pid_attitude', 'pitch_ki'))
        mappers.update(self._create_slider(self.grid_attitude, 2, 3, 0, 10, 'pid_attitude', 'pitch_kd'))

        mappers.update(self._create_slider(self.grid_attitude, 3, 1, 0, 10, 'pid_attitude', 'yaw_kp'))
        mappers.update(self._create_slider(self.grid_attitude, 3, 2, 0, 10, 'pid_attitude', 'yaw_ki'))
        mappers.update(self._create_slider(self.grid_attitude, 3, 3, 0, 10, 'pid_attitude', 'yaw_kd'))

        self._link(mappers, 'pid_attitude.roll_kp', 'pid_attitude.pitch_kp', self.attitude_link_checkbox)
        self._link(mappers, 'pid_attitude.roll_ki', 'pid_attitude.pitch_ki', self.attitude_link_checkbox)
        self._link(mappers, 'pid_attitude.roll_kd', 'pid_attitude.pitch_kd', self.attitude_link_checkbox)

        # Position control PID
        mappers.update(self._create_slider(self.grid_pos_ctrl, 1, 1, 0, 10, 'posCtlPid', 'xKp'))
        mappers.update(self._create_slider(self.grid_pos_ctrl, 1, 2, 0, 10, 'posCtlPid', 'xKi'))
        mappers.update(self._create_slider(self.grid_pos_ctrl, 1, 3, 0, 10, 'posCtlPid', 'xKd'))

        mappers.update(self._create_slider(self.grid_pos_ctrl, 2, 1, 0, 10, 'posCtlPid', 'yKp'))
        mappers.update(self._create_slider(self.grid_pos_ctrl, 2, 2, 0, 10, 'posCtlPid', 'yKi'))
        mappers.update(self._create_slider(self.grid_pos_ctrl, 2, 3, 0, 10, 'posCtlPid', 'yKd'))

        mappers.update(self._create_slider(self.grid_pos_ctrl, 3, 1, 0, 10, 'posCtlPid', 'zKp'))
        mappers.update(self._create_slider(self.grid_pos_ctrl, 3, 2, 0, 10, 'posCtlPid', 'zKi'))
        mappers.update(self._create_slider(self.grid_pos_ctrl, 3, 3, 0, 10, 'posCtlPid', 'zKd'))

        mappers.update(self._create_slider(self.grid_pos_ctrl, 5, 1, 0, 65536, 'posCtlPid', 'thrustBase'))

        self._link(mappers, 'posCtlPid.xKp', 'posCtlPid.yKp', self.position_link_checkbox)
        self._link(mappers, 'posCtlPid.xKi', 'posCtlPid.yKi', self.position_link_checkbox)
        self._link(mappers, 'posCtlPid.xKd', 'posCtlPid.yKd', self.position_link_checkbox)

        # Velocity control PID
        mappers.update(self._create_slider(self.grid_vel_ctrl, 1, 1, 0, 50, 'velCtlPid', 'vxKp'))
        mappers.update(self._create_slider(self.grid_vel_ctrl, 1, 2, 0, 50, 'velCtlPid', 'vxKi'))
        mappers.update(self._create_slider(self.grid_vel_ctrl, 1, 3, 0, 10, 'velCtlPid', 'vxKd'))
        mappers.update(self._create_slider(self.grid_vel_ctrl, 1, 4, 0, 10, 'velCtlPid', 'vxKFF'))

        mappers.update(self._create_slider(self.grid_vel_ctrl, 2, 1, 0, 50, 'velCtlPid', 'vyKp'))
        mappers.update(self._create_slider(self.grid_vel_ctrl, 2, 2, 0, 50, 'velCtlPid', 'vyKi'))
        mappers.update(self._create_slider(self.grid_vel_ctrl, 2, 3, 0, 10, 'velCtlPid', 'vyKd'))
        mappers.update(self._create_slider(self.grid_vel_ctrl, 2, 4, 0, 10, 'velCtlPid', 'vyKFF'))

        mappers.update(self._create_slider(self.grid_vel_ctrl, 3, 1, 0, 50, 'velCtlPid', 'vzKp'))
        mappers.update(self._create_slider(self.grid_vel_ctrl, 3, 2, 0, 50, 'velCtlPid', 'vzKi'))
        mappers.update(self._create_slider(self.grid_vel_ctrl, 3, 3, 0, 10, 'velCtlPid', 'vzKd'))

        self._link(mappers, 'velCtlPid.vxKp', 'velCtlPid.vyKp', self.velocity_link_checkbox)
        self._link(mappers, 'velCtlPid.vxKi', 'velCtlPid.vyKi', self.velocity_link_checkbox)
        self._link(mappers, 'velCtlPid.vxKd', 'velCtlPid.vyKd', self.velocity_link_checkbox)
        self._link(mappers, 'velCtlPid.vxKFF', 'velCtlPid.vyKFF', self.velocity_link_checkbox)

        return mappers

    def _connected(self, link_uri):
        """Callback when the Crazyflie has been connected"""
        for mapper in self.mappers.values():
            param_group, param_name = mapper.connected(self._helper.cf)
            self._helper.cf.param.add_update_callback(
                group=param_group, name=param_name, cb=self._param_updated_signal.emit)

        self._enable_ui_objects(True)

    def _disconnected(self, link_uri):
        """Callback for when the Crazyflie has been disconnected"""
        for mapper in self.mappers.values():
            mapper.disconnected()

        self._enable_ui_objects(False)
        # TODO set default values?

    def _create_slider(self, gridLayout, row, col, min_val, max_val, param_group: str, param_name: str):
        initial_val = (min_val + max_val) / 2.0
        slider = SuperSlider(min_val, max_val, initial_val)
        gridLayout.addWidget(slider, row, col)

        slider_mapper = SliderParamMapper(slider, param_group, param_name)

        return {slider_mapper.full_param_name: slider_mapper}

    def _param_updated_cb(self, full_param_name, value):
        if full_param_name in self.mappers:
            self.mappers[full_param_name].param_updated_cb(full_param_name, value)

    def _link(self, mappers, first, other, checkbox):
        first_mapper = mappers[first]
        other_mapper = mappers[other]

        first_mapper.link_with(other_mapper, checkbox)

    def _store_button_clicked(self):
        param: Param = self._helper.cf.param
        for full_param_name in self.mappers.keys():
            # syncer = Syncer()
            param.persistent_store(full_param_name)

    def _clear_stored_button_clicked(self):
        param: Param = self._helper.cf.param
        for full_param_name in self.mappers.keys():
            param.persistent_clear(full_param_name)

    def _default_values_button_clicked(self):
        param: Param = self._helper.cf.param
        for full_param_name in self.mappers.keys():
            # For some reason we run into problems if we try to get all params in one go. Pace by getting them
            # one by one.
            syncer = Syncer()
            param.get_default_value(full_param_name, syncer.success_cb)
            syncer.wait()
            self._param_updated_cb(*syncer.success_args)

    def _enable_ui_objects(self, enabled):
        objects = [
            self.rate_link_checkbox,
            self.attitude_link_checkbox,
            self.position_link_checkbox,
            self.velocity_link_checkbox,
            self.store_button,
            self.clear_stored_button,
            self.default_values_button,
        ]

        for item in objects:
            item.setEnabled(enabled)

        for mapper in self.mappers.values():
            mapper.enable_ui(enabled)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
This tab plots different logging data defined by configurations that has been
pre-configured.
"""

import logging

from cfclient.ui.tab_toolbox import TabToolbox
from cfclient.ui.widgets.plotwidget import PlotWidget
from PyQt6 import uic
from PyQt6.QtCore import pyqtSignal
from PyQt6.QtCore import QAbstractItemModel
from PyQt6.QtCore import QModelIndex
from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QMessageBox

import cfclient

__author__ = 'Bitcraze AB'
__all__ = ['PlotTab']

logger = logging.getLogger(__name__)

plot_tab_class = uic.loadUiType(cfclient.module_path + "/ui/tabs/plotTab.ui")[0]


class LogConfigModel(QAbstractItemModel):
    """Model for log configurations in the ComboBox"""

    def __init__(self, parent=None):
        super(LogConfigModel, self).__init__(parent)
        self._nodes = []

    def add_block(self, block):
        self._nodes.append(block)
        self.layoutChanged.emit()
        self._nodes.sort(key=lambda conf: conf.name.lower())

    def parent(self, index):
        """Re-implemented method to get the parent of the given index"""
        return QModelIndex()

    def remove_block(self, block):
        """Remove a block from the view"""
        self._nodes.remove(block)

    def columnCount(self, parent):
        """Re-implemented method to get the number of columns"""
        return 1

    def rowCount(self, parent):
        """Re-implemented method to get the number of rows for a given index"""
        parent_item = parent.internalPointer()
        if parent.isValid():
            parent_item = parent.internalPointer()  # noqa
            return 0
        else:
            return len(self._nodes)

    def index(self, row, column, parent):
        """Re-implemented method to get the index for a specified
        row/column/parent combination"""
        if not self._nodes:
            return QModelIndex()
        node = parent.internalPointer()
        if not node:
            index = self.createIndex(row, column, self._nodes[row])
            return index
        else:
            return self.createIndex(row, column, node.get_child(row))

    def data(self, index, role):
        """Re-implemented method to get the data for a given index and role"""
        node = index.internalPointer()  # noqa
        if not index.isValid() or not 0 <= index.row() < len(self._nodes):
            return None
        if role == Qt.ItemDataRole.DisplayRole:
            return self._nodes[index.row()].name
        return None

    def reset(self):
        """Reset the model"""
        self._nodes = []
        self.layoutChanged.emit()

    def get_config(self, i):
        return self._nodes[i]


class PlotTab(TabToolbox, plot_tab_class):
    """Tab for plotting logging data"""

    _log_data_signal = pyqtSignal(int, object, object)
    _log_error_signal = pyqtSignal(object, str)
    _disconnected_signal = pyqtSignal(str)
    _connected_signal = pyqtSignal(str)

    colors = [
        (60, 200, 60),    # green
        (40, 100, 255),   # blue
        (255, 130, 240),  # magenta
        (255, 26, 28),    # red
        (255, 170, 0),    # orange
        (40, 180, 240),   # cyan
        (153, 153, 153),  # grey
        (176, 96, 50),    # brown
        (180, 60, 240),   # purple
    ]

    def __init__(self, helper):
        super(PlotTab, self).__init__(helper, 'Plotter')
        self.setupUi(self)

        self._log_error_signal.connect(self._logging_error)

        self._plot = PlotWidget(fps=30)
        # Check if we could find the PyQtImport. If not, then
        # set this tab as disabled
        is_enabled = self._plot.can_enable

        self._model = LogConfigModel()
        self.dataSelector.setModel(self._model)
        self._log_data_signal.connect(self._log_data_received)
        self.plotLayout.addWidget(self._plot)

        # Connect external signals if we can use the tab
        if is_enabled:
            self._disconnected_signal.connect(self._disconnected)
            self._helper.cf.disconnected.add_callback(
                self._disconnected_signal.emit)

            self._connected_signal.connect(self._connected)
            self._helper.cf.connected.add_callback(
                self._connected_signal.emit)

            self._helper.cf.log.block_added_cb.add_callback(self._config_added)
            self.dataSelector.currentIndexChanged.connect(
                self._selection_changed)

        self._previous_config = None
        self._started_previous = False

    def _connected(self, link_uri):
        """Callback when the Crazyflie has been connected"""
        self._plot.removeAllDatasets()
        self._plot.set_title("")

    def _disconnected(self, link_uri):
        """Callback for when the Crazyflie has been disconnected"""
        self._model.beginResetModel()
        self._model.reset()
        self._model.endResetModel()
        self.dataSelector.setCurrentIndex(-1)
        self._previous_config = None
        self._started_previous = False

    def _log_data_signal_wrapper(self, ts, data, logconf):
        """Wrapper for signal"""

        # For some reason the *.emit functions are not
        # the same over time (?!) so they cannot be registered and then
        # removed as callbacks.
        self._log_data_signal.emit(ts, data, logconf)

    def _log_error_signal_wrapper(self, config, msg):
        """Wrapper for signal"""

        # For some reason the *.emit functions are not
        # the same over time (?!) so they cannot be registered and then
        # removed as callbacks.
        self._log_error_signal.emit(config, msg)

    def _selection_changed(self, i):
        """Callback from ComboBox when a new item has been selected"""

        # Check if we have disconnected
        if i < 0:
            return
        # First check if we need to stop the old block
        if self._started_previous and self._previous_config:
            logger.debug("Should stop config [%s], stopping!",
                         self._previous_config.name)
            self._previous_config.delete()

        # Remove our callback for the previous config
        if self._previous_config:
            self._previous_config.data_received_cb.remove_callback(
                self._log_data_signal_wrapper)
            self._previous_config.error_cb.remove_callback(
                self._log_error_signal_wrapper)

        lg = self._model.get_config(i)
        if not lg.started:
            logger.debug("Config [%s] not started, starting!", lg.name)
            self._started_previous = True
            lg.start()
        else:
            self._started_previous = False
        self._plot.removeAllDatasets()
        color_selector = 0

        self._plot.set_title(lg.name)

        for d in lg.variables:
            self._plot.add_curve(d.name, self.colors[
                color_selector % len(self.colors)])
            color_selector += 1
        lg.data_received_cb.add_callback(self._log_data_signal_wrapper)
        lg.error_cb.add_callback(self._log_error_signal_wrapper)

        self._previous_config = lg

    def _config_added(self, logconfig):
        """Callback from the log layer when a new config has been added"""
        logger.debug("Callback for new config [%s]", logconfig.name)
        self._model.add_block(logconfig)

    def remove_config(self, logconfig):
        self._model.remove_block(logconfig)

    def _logging_error(self, log_conf, msg):
        """Callback from the log layer when an error occurs"""
        QMessageBox.about(
            self, "Plot error", "Error when starting log config [%s]: %s" % (
                log_conf.name, msg))

    def _log_data_received(self, timestamp, data, logconf):
        """Callback when the log layer receives new data"""

        # Check so that the incoming data belongs to what we are currently
        # logging
        if self._previous_config:
            if self._previous_config.name == logconf.name:
                self._plot.add_data(data, timestamp)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2011-2023 Bitcraze AB
#
#  Crazyflie Nano Quadcopter Client
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License along with
#  this program; if not, write to the Free Software Foundation, Inc., 51
#  Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
The main file for the Crazyflie control application.
"""
import logging
import sys
import usb

import cfclient
from cfclient.ui.pose_logger import PoseLogger
from cfclient.ui.tab_toolbox import TabToolbox
import cfclient.ui.tabs
import cflib.crtp
from cfclient.ui.dialogs.about import AboutDialog
from cfclient.ui.dialogs.bootloader import BootloaderDialog
from cfclient.ui.connectivity_manager import ConnectivityManager
from cfclient.utils.config import Config
from cfclient.utils.config_manager import ConfigManager
from cfclient.utils.input import JoystickReader
from cfclient.utils.logconfigreader import LogConfigReader
from cfclient.utils.ui import UiUtils
from cfclient.utils.zmq_led_driver import ZMQLEDDriver
from cfclient.utils.zmq_param import ZMQParamAccess
from cflib.crazyflie import Crazyflie
from cflib.crazyflie.log import LogConfig
from cflib.crazyflie.mem import MemoryElement
from PyQt6 import QtWidgets
from PyQt6 import uic
from PyQt6.QtCore import pyqtSignal
from PyQt6.QtCore import pyqtSlot
from PyQt6.QtCore import QDir
from PyQt6.QtCore import QThread
from PyQt6.QtCore import QUrl
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QAction
from PyQt6.QtGui import QActionGroup
from PyQt6.QtGui import QShortcut
from PyQt6.QtGui import QDesktopServices
from PyQt6.QtGui import QPalette
from PyQt6.QtWidgets import QLabel
from PyQt6.QtWidgets import QMenu
from PyQt6.QtWidgets import QMessageBox

from .dialogs.cf2config import Cf2ConfigDialog
from .dialogs.inputconfigdialogue import InputConfigDialogue
from .dialogs.logconfigdialogue import LogConfigDialogue


__author__ = 'Bitcraze AB'
__all__ = ['MainUI']

logger = logging.getLogger(__name__)

(main_window_class,
 main_windows_base_class) = (uic.loadUiType(cfclient.module_path +
                                            '/ui/main.ui'))


class UIState:
    DISCONNECTED = 0
    CONNECTING = 1
    CONNECTED = 2
    SCANNING = 3


class BatteryStates:
    BATTERY, CHARGING, CHARGED, LOW_POWER = list(range(4))


class MainUI(QtWidgets.QMainWindow, main_window_class):
    connectionLostSignal = pyqtSignal(str, str)
    connectionInitiatedSignal = pyqtSignal(str)
    batteryUpdatedSignal = pyqtSignal(int, object, object)
    connectionDoneSignal = pyqtSignal(str)
    connectionFailedSignal = pyqtSignal(str, str)
    disconnectedSignal = pyqtSignal(str)
    linkQualitySignal = pyqtSignal(float)

    _input_device_error_signal = pyqtSignal(str)
    _input_discovery_signal = pyqtSignal(object)
    _log_error_signal = pyqtSignal(object, str)

    def __init__(self, *args):
        super(MainUI, self).__init__(*args)
        self.setupUi(self)

        # Restore window size if present in the config file
        try:
            size = Config().get("window_size")
            self.resize(size[0], size[1])
        except KeyError:
            pass

        self.cf = Crazyflie(ro_cache=None,
                            rw_cache=cfclient.config_path + "/cache")

        cflib.crtp.init_drivers()

        zmq_params = ZMQParamAccess(self.cf)
        zmq_params.start()

        zmq_leds = ZMQLEDDriver(self.cf)
        zmq_leds.start()

        self.scanner = ScannerThread()
        self.scanner.interfaceFoundSignal.connect(self.foundInterfaces)
        self.scanner.start()

        # Create and start the Input Reader
        self._statusbar_label = QLabel("No input-device found, insert one to"
                                       " fly.")
        self.statusBar().addWidget(self._statusbar_label)

        #
        # We use this hacky-trick to find out if we are in dark-mode and
        # figure out what bgcolor to set from that. We always use the current
        # palette forgreound.
        #
        self.textColor = self._statusbar_label.palette().color(QPalette.ColorRole.WindowText)
        self.bgColor = self._statusbar_label.palette().color(QPalette.ColorRole.Window)
        self.isDark = self.textColor.value() > self.bgColor.value()

        self.joystickReader = JoystickReader()
        self._active_device = ""
        # self.configGroup = QActionGroup(self._menu_mappings, exclusive=True)

        self._mux_group = QActionGroup(self._menu_inputdevice)
        self._mux_group.setExclusive(True)

        # TODO: Need to reload configs
        # ConfigManager().conf_needs_reload.add_callback(self._reload_configs)

        self.connect_input = QShortcut("Ctrl+I", self.connectButton, self._connect)
        self.cf.connection_failed.add_callback(
            self.connectionFailedSignal.emit)
        self.connectionFailedSignal.connect(self._connection_failed)

        self._input_device_error_signal.connect(
            self._display_input_device_error)
        self.joystickReader.device_error.add_callback(
            self._input_device_error_signal.emit)
        self._input_discovery_signal.connect(self.device_discovery)
        self.joystickReader.device_discovery.add_callback(
            self._input_discovery_signal.emit)

        # Connect UI signals
        self.logConfigAction.triggered.connect(self._show_connect_dialog)
        self.menuItemConnect.triggered.connect(self._connect)
        self.menuItemConfInputDevice.triggered.connect(
            self._show_input_device_config_dialog)
        self.menuItemExit.triggered.connect(self.closeAppRequest)
        self.batteryUpdatedSignal.connect(self._update_battery)
        self._menuitem_rescandevices.triggered.connect(self._rescan_devices)
        self._menuItem_openconfigfolder.triggered.connect(
            self._open_config_folder)

        self._set_address()

        self._connectivity_manager = ConnectivityManager()
        self._connectivity_manager.register_ui_elements(
            ConnectivityManager.UiElementsContainer(
                interface_combo=self.interfaceCombo,
                address_spinner=self.address,
                connect_button=self.connectButton,
                scan_button=self.scanButton))

        self._connectivity_manager.connect_button_clicked.connect(self._connect)
        self._connectivity_manager.scan_button_clicked.connect(self._scan_from_button)

        self._disable_input = False

        self.joystickReader.input_updated.add_callback(
            lambda *args: self._disable_input or
            self.cf.commander.send_setpoint(*args))

        self.joystickReader.assisted_input_updated.add_callback(
            lambda *args: self._disable_input or
            self.cf.commander.send_velocity_world_setpoint(*args))

        self.joystickReader.heighthold_input_updated.add_callback(
            lambda *args: self._disable_input or
            self.cf.commander.send_zdistance_setpoint(*args))

        self.joystickReader.hover_input_updated.add_callback(
            self.cf.commander.send_hover_setpoint)

        # Emergency stop button
        self.esButton.clicked.connect(self._emergency_stop)

        # Connection callbacks and signal wrappers for UI protection
        self.cf.connected.add_callback(self.connectionDoneSignal.emit)
        self.connectionDoneSignal.connect(self._connected)
        self.cf.disconnected.add_callback(self.disconnectedSignal.emit)
        self.disconnectedSignal.connect(self._disconnected)
        self.cf.connection_lost.add_callback(self.connectionLostSignal.emit)
        self.connectionLostSignal.connect(self._connection_lost)
        self.cf.connection_requested.add_callback(
            self.connectionInitiatedSignal.emit)
        self.connectionInitiatedSignal.connect(self._connection_initiated)
        self._log_error_signal.connect(self._logging_error)

        self.batteryBar.setTextVisible(False)
        self.linkQualityBar.setTextVisible(False)

        # Connect link quality feedback
        self.cf.link_statistics.link_quality_updated.add_callback(self.linkQualitySignal.emit)
        self.linkQualitySignal.connect(
            lambda percentage: self.linkQualityBar.setValue(int(percentage)))

        # Parse the log configuration files
        self.logConfigReader = LogConfigReader(self.cf)

        self._current_input_config = None
        self._active_config = None
        self._active_config = None

        self.inputConfig = None

        # Add things to helper so tabs can access it
        cfclient.ui.pluginhelper.cf = self.cf
        cfclient.ui.pluginhelper.inputDeviceReader = self.joystickReader
        cfclient.ui.pluginhelper.logConfigReader = self.logConfigReader
        cfclient.ui.pluginhelper.pose_logger = PoseLogger(self.cf)
        cfclient.ui.pluginhelper.connectivity_manager = self._connectivity_manager
        cfclient.ui.pluginhelper.mainUI = self

        self.logConfigDialogue = LogConfigDialogue(cfclient.ui.pluginhelper)
        self._bootloader_dialog = BootloaderDialog(cfclient.ui.pluginhelper)
        self._cf2config_dialog = Cf2ConfigDialog(cfclient.ui.pluginhelper)
        self.menuItemBootloader.triggered.connect(self._bootloader_dialog.show)
        self._about_dialog = AboutDialog(cfclient.ui.pluginhelper)
        self.menuItemAbout.triggered.connect(self._about_dialog.show)
        self._menu_cf2_config.triggered.connect(self._cf2config_dialog.show)

        self._connectivity_manager.set_address(self.address.value())

        self._initial_scan = True
        self._scan(self._connectivity_manager.get_address())

        self.tabs_menu_item = QMenu("Tabs", self.menuView, enabled=True)
        self.menuView.addMenu(self.tabs_menu_item)

        self.toolboxes_menu_item = QMenu("Toolboxes", self.menuView, enabled=True)
        self.menuView.addMenu(self.toolboxes_menu_item)

        self.loaded_tab_toolboxes = self.create_tab_toolboxes(self.tabs_menu_item,
                                                              self.toolboxes_menu_item,
                                                              self.tab_widget)
        self.read_tab_toolbox_config(self.loaded_tab_toolboxes)

        # References to all the device sub-menus in the "Input device" menu
        self._all_role_menus = ()
        # Used to filter what new devices to add default mapping to
        self._available_devices = ()
        # Keep track of mux nodes so we can enable according to how many
        # devices we have
        self._all_mux_nodes = ()

        # Check which Input muxes are available
        self._mux_group = QActionGroup(self._menu_inputdevice)
        self._mux_group.setExclusive(True)
        for m in self.joystickReader.available_mux():
            node = QAction(m.name,
                           self._menu_inputdevice,
                           checkable=True,
                           enabled=False)
            node.toggled.connect(self._mux_selected)
            self._mux_group.addAction(node)
            self._menu_inputdevice.addAction(node)
            self._all_mux_nodes += (node,)
            mux_subnodes = ()
            for name in m.supported_roles():
                sub_node = QMenu("    {}".format(name),
                                 self._menu_inputdevice,
                                 enabled=False)
                self._menu_inputdevice.addMenu(sub_node)
                mux_subnodes += (sub_node,)
                self._all_role_menus += ({"muxmenu": node,
                                          "rolemenu": sub_node},)
            node.setData((m, mux_subnodes))

        self._mapping_support = True

        # Add checkbuttons for theme-selection.
        self._theme_group = QActionGroup(self.menuThemes)
        self._theme_group.setExclusive(True)
        self._theme_checkboxes = []
        for theme in UiUtils.THEMES:
            node = QAction(theme, self.menuThemes, checkable=True)
            node.setObjectName(theme)
            node.toggled.connect(self._theme_selected)
            self._theme_checkboxes.append(node)
            self._theme_group.addAction(node)
            self.menuThemes.addAction(node)

        # We only want to warn about USB permission once
        self._permission_warned = False

    def create_tab_toolboxes(self, tabs_menu_item, toolboxes_menu_item, tab_widget):
        loaded_tab_toolboxes = {}

        for tab_class in cfclient.ui.tabs.available:
            tab_toolbox = tab_class(cfclient.ui.pluginhelper)
            loaded_tab_toolboxes[tab_toolbox.get_tab_toolbox_name()] = tab_toolbox

            # Set reference for plot-tab.
            if isinstance(tab_toolbox, cfclient.ui.tabs.PlotTab):
                cfclient.ui.pluginhelper.plotTab = tab_toolbox

            # Add to tabs menu
            tab_action_item = QAction(tab_toolbox.get_tab_toolbox_name())
            tab_action_item.setCheckable(True)
            tab_action_item.triggered.connect(self.toggle_tab_visibility)
            tab_action_item.tab_toolbox = tab_toolbox
            tab_toolbox.tab_action_item = tab_action_item

            tabs_menu_item.addAction(tab_action_item)

            # Add to toolbox menu
            toolbox_action_item = QAction(tab_toolbox.get_tab_toolbox_name())
            toolbox_action_item.setCheckable(True)
            toolbox_action_item.triggered.connect(self.toggle_toolbox_visibility)
            toolbox_action_item.tab_toolbox = tab_toolbox
            tab_toolbox.toolbox_action_item = toolbox_action_item
            tab_toolbox.dock_widget.closed.connect(lambda: self.toggle_toolbox_visibility(False))
            tab_toolbox.dock_widget.dockLocationChanged.connect(lambda area: self.set_preferred_dock_area(area))

            toolboxes_menu_item.addAction(toolbox_action_item)

        return loaded_tab_toolboxes

    def read_tab_toolbox_config(self, loaded_tab_toolboxes):
        # Add tabs in the correct order
        for name in TabToolbox.read_open_tab_config():
            if name in loaded_tab_toolboxes.keys():
                self._tab_toolbox_show_as_tab(loaded_tab_toolboxes[name])

        for name in TabToolbox.read_open_toolbox_config():
            if name in loaded_tab_toolboxes.keys():
                self._tab_toolbox_show_as_toolbox(loaded_tab_toolboxes[name])

    def _set_address(self):
        address = 0xE7E7E7E7E7
        try:
            link_uri = Config().get("link_uri")
            if link_uri.startswith("radio://"):
                if len(link_uri) > 0:
                    parts = link_uri.split('/')
                    # The uri might not contain an address
                    if len(parts) == 6:
                        address = int(parts[-1], 16)
        except Exception as err:
            logger.warn('failed to parse address from config: %s' % str(err))
        finally:
            self.address.setValue(address)

    def _theme_selected(self, *args):
        """ Callback when a theme is selected. """
        for checkbox in self._theme_checkboxes:
            if checkbox.isChecked():
                theme = checkbox.objectName()
                app = QtWidgets.QApplication.instance()
                app.setStyleSheet(UiUtils.select_theme(theme))
                Config().set('theme', theme)

    def _check_theme(self, theme_name):
        # Check the default theme.
        for theme in self._theme_checkboxes:
            if theme.objectName() == theme_name:
                theme.setChecked(True)
                self._theme_selected(True)

    def set_default_theme(self):
        try:
            theme = Config().get('theme')
        except KeyError:
            theme = 'Default'
        self._check_theme(theme)

    def disable_input(self, disable):
        """
        Disable the gamepad input to be able to send setpoint from a tab
        """
        self._disable_input = disable

    def foundInterfaces(self, interfaces):
        selected_interface = self._connectivity_manager.get_interface()

        formatted_interfaces = []
        for i in interfaces:
            if len(i[1]) > 0:
                interface = "%s - %s" % (i[0], i[1])
            else:
                interface = i[0]
            formatted_interfaces.append(interface)

        if self._initial_scan:
            self._initial_scan = False

            try:
                if len(Config().get("link_uri")) > 0:
                    formatted_interfaces.index(Config().get("link_uri"))
                    selected_interface = Config().get("link_uri")
            except KeyError:
                #  The configuration for link_uri was not found
                pass
            except ValueError:
                #  The saved URI was not found while scanning
                pass

        if len(interfaces) == 1 and selected_interface is None:
            selected_interface = interfaces[0][0]

        newIndex = None
        if selected_interface is not None:
            try:
                newIndex = formatted_interfaces.index(selected_interface)
            except ValueError:
                pass

        self._connectivity_manager.set_interfaces(formatted_interfaces, newIndex)

        self.uiState = UIState.DISCONNECTED
        self._update_ui_state()

    def _update_ui_state(self):
        if self.uiState == UIState.DISCONNECTED:
            self.setWindowTitle("Not connected")
            canConnect = self._connectivity_manager.get_interface() is not None
            self.menuItemConnect.setText("Connect to Crazyflie")
            self.menuItemConnect.setEnabled(canConnect)
            self._connectivity_manager.set_state(ConnectivityManager.UIState.DISCONNECTED)
            self.batteryBar.setValue(3000)
            self._menu_cf2_config.setEnabled(False)
            self.linkQualityBar.setValue(0)
            self.logConfigAction.setEnabled(False)
            self.esButton.setStyleSheet("")
            self.esButton.setEnabled(False)
        elif self.uiState == UIState.CONNECTED:
            s = "Connected on %s" % self._connectivity_manager.get_interface()
            self.setWindowTitle(s)
            self.menuItemConnect.setText("Disconnect")
            self.menuItemConnect.setEnabled(True)
            self._connectivity_manager.set_state(ConnectivityManager.UIState.CONNECTED)
            self.logConfigAction.setEnabled(True)
            self.esButton.setEnabled(True)
            self.esButton.setStyleSheet("background-color: red")
            # Find out if there's an I2C EEPROM, otherwise don't show the
            # dialog.
            if len(self.cf.mem.get_mems(MemoryElement.TYPE_I2C)) > 0:
                self._menu_cf2_config.setEnabled(True)
        elif self.uiState == UIState.CONNECTING:
            s = "Connecting to {} ...".format(self._connectivity_manager.get_interface())
            self.setWindowTitle(s)
            self.menuItemConnect.setText("Cancel")
            self.menuItemConnect.setEnabled(True)
            self._connectivity_manager.set_state(ConnectivityManager.UIState.CONNECTING)
        elif self.uiState == UIState.SCANNING:
            self.setWindowTitle("Scanning ...")
            self.menuItemConnect.setEnabled(False)
            self._connectivity_manager.set_state(ConnectivityManager.UIState.SCANNING)

    @pyqtSlot(bool)
    def toggle_tab_visibility(self, checked):
        tab_action_item = self.sender()
        tab_toolbox = tab_action_item.tab_toolbox

        if checked:
            self._tab_toolbox_show_as_tab(tab_toolbox)
        else:
            self._tab_toolbox_hide(tab_toolbox)

    @pyqtSlot(bool)
    def toggle_toolbox_visibility(self, checked):
        toolbox_action_item = self.sender()
        tab_toolbox = toolbox_action_item.tab_toolbox

        if checked:
            self._tab_toolbox_show_as_toolbox(tab_toolbox)
        else:
            self._tab_toolbox_hide(tab_toolbox)

    def _tab_toolbox_show_as_tab(self, tab_toolbox):
        if tab_toolbox.get_display_state() == TabToolbox.DS_TOOLBOX:
            dock_widget = tab_toolbox.dock_widget
            self.removeDockWidget(dock_widget)
            dock_widget.hide()

        if tab_toolbox.get_display_state() != TabToolbox.DS_TAB:
            tab_toolbox_name = tab_toolbox.get_tab_toolbox_name()
            self.tab_widget.addTab(tab_toolbox, tab_toolbox_name)

        tab_toolbox.tab_action_item.setChecked(True)
        tab_toolbox.toolbox_action_item.setChecked(False)
        tab_toolbox.set_display_state(TabToolbox.DS_TAB)

    def _tab_toolbox_show_as_toolbox(self, tab_toolbox):
        dock_widget = tab_toolbox.dock_widget

        if tab_toolbox.get_display_state() == TabToolbox.DS_TAB:
            self.tab_widget.removeTab(self.tab_widget.indexOf(tab_toolbox))

        if tab_toolbox.get_display_state() != TabToolbox.DS_TOOLBOX:
            self.addDockWidget(tab_toolbox.preferred_dock_area(), dock_widget)
            dock_widget.setWidget(tab_toolbox)
            dock_widget.show()

        tab_toolbox.tab_action_item.setChecked(False)
        tab_toolbox.toolbox_action_item.setChecked(True)
        tab_toolbox.set_display_state(TabToolbox.DS_TOOLBOX)

    def _tab_toolbox_hide(self, tab_toolbox):
        dock_widget = tab_toolbox.dock_widget

        if tab_toolbox.get_display_state() == TabToolbox.DS_TAB:
            self.tab_widget.removeTab(self.tab_widget.indexOf(tab_toolbox))
        elif tab_toolbox.get_display_state() == TabToolbox.DS_TOOLBOX:
            self.removeDockWidget(dock_widget)
            dock_widget.hide()
            tab_toolbox.toolbox_action_item.setChecked(False)

        tab_toolbox.tab_action_item.setChecked(False)
        tab_toolbox.toolbox_action_item.setChecked(False)
        tab_toolbox.set_display_state(TabToolbox.DS_HIDDEN)

    @pyqtSlot(Qt.DockWidgetArea)
    def set_preferred_dock_area(self, area):
        dock_widget = self.sender()
        tab_toolbox = dock_widget.tab_toolbox
        tab_toolbox.set_preferred_dock_area(area)

    def _rescan_devices(self):
        self._statusbar_label.setText("No inputdevice connected!")
        self._menu_devices.clear()
        self._active_device = ""
        self.joystickReader.stop_input()

        # for c in self._menu_mappings.actions():
        #    c.setEnabled(False)
        # devs = self.joystickReader.available_devices()
        # if (len(devs) > 0):
        #    self.device_discovery(devs)

    def _show_input_device_config_dialog(self):
        self.inputConfig = InputConfigDialogue(self.joystickReader)
        self.inputConfig.show()

    def _show_connect_dialog(self):
        self.logConfigDialogue.show()

    def _emergency_stop(self):
        # send disarming command
        if (self.uiState == UIState.CONNECTED):
            # Send both emergency stop and disarm
            # TODO krri Disarm?
            self.cf.loc.send_emergency_stop()

    def _update_battery(self, timestamp, data, logconf):
        self.batteryBar.setValue(int(data["pm.vbat"] * 1000))

        color = UiUtils.COLOR_BLUE
        # TODO firmware reports fully-charged state as 'Battery',
        # rather than 'Charged'
        if data["pm.state"] in [BatteryStates.CHARGING, BatteryStates.CHARGED]:
            color = UiUtils.COLOR_GREEN
        elif data["pm.state"] == BatteryStates.LOW_POWER:
            color = UiUtils.COLOR_RED

        self.batteryBar.setStyleSheet(UiUtils.progressbar_stylesheet(color))
        self._aff_volts.setText(("%.3f" % data["pm.vbat"]))

    def _connected(self):
        self.uiState = UIState.CONNECTED
        self._update_ui_state()

        Config().set("link_uri", str(self._connectivity_manager.get_interface()))

        lg = LogConfig("Battery", 1000)
        lg.add_variable("pm.vbat", "float")
        lg.add_variable("pm.state", "int8_t")
        try:
            self.cf.log.add_config(lg)
            lg.data_received_cb.add_callback(self.batteryUpdatedSignal.emit)
            lg.error_cb.add_callback(self._log_error_signal.emit)
            lg.start()
        except KeyError as e:
            logger.warning(str(e))

        mems = self.cf.mem.get_mems(MemoryElement.TYPE_DRIVER_LED)
        if len(mems) > 0:
            mems[0].write_data(self._led_write_done)

    def _disconnected(self):
        self.uiState = UIState.DISCONNECTED
        self._update_ui_state()

    def _connection_initiated(self):
        self.uiState = UIState.CONNECTING
        self._update_ui_state()

    def _led_write_done(self, mem, addr):
        logger.info("LED write done callback")

    def _logging_error(self, log_conf, msg):
        QMessageBox.about(self, "Log error", "Error when starting log config"
                                             " [{}]: {}".format(log_conf.name,
                                                                msg))

    def _connection_lost(self, linkURI, msg):
        if self.isActiveWindow():
            warningCaption = "Communication failure"
            error = "Connection lost to {}: {}".format(linkURI, msg)
            QMessageBox.critical(self, warningCaption, error)
            self.uiState = UIState.DISCONNECTED
            self._update_ui_state()

    def _connection_failed(self, linkURI, error):
        msg = "Failed to connect on {}: {}".format(linkURI, error)
        warningCaption = "Communication failure"
        QMessageBox.critical(self, warningCaption, msg)
        self.uiState = UIState.DISCONNECTED
        self._update_ui_state()

    def closeEvent(self, event):
        Config().save_file()
        self.cf.close_link()
        self.hide()

    def resizeEvent(self, event):
        Config().set("window_size", [event.size().width(),
                                     event.size().height()])

    def _connect(self):
        if self.uiState == UIState.CONNECTED:
            self.cf.close_link()
        elif self.uiState == UIState.CONNECTING:
            self.cf.close_link()
            self.uiState = UIState.DISCONNECTED
            self._update_ui_state()
        else:
            self.cf.open_link(self._connectivity_manager.get_interface())

    def _scan(self, address):
        self.uiState = UIState.SCANNING
        self._update_ui_state()
        self.scanner.scanSignal.emit(address)

    def _scan_from_button(self, address):
        #
        # Below we check if we can open the Crazyradio device.
        # If it is there, but we have no permissions we inform the user, once,
        # about how to install the udev rules.
        #
        if not self._permission_warned:
            try:
                radio = cflib.crtp.radiodriver.RadioManager.open(0)
                radio.close()
            except usb.core.USBError as e:
                if e.errno == 13:  # Permission denied
                    link = "<a href='https://www.bitcraze.io/documentation/repository/crazyflie-lib-python/master/installation/usb_permissions/'>Install USB Permissions</a>" # noqa
                    msg = QMessageBox()
                    msg.setIcon(QMessageBox.Icon.Information)
                    msg.setTextFormat(Qt.TextFormat.RichText)
                    msg.setText("Could not access Crazyradio")
                    msg.setInformativeText(link)
                    msg.setWindowTitle("Crazyradio permissions")
                    msg.exec()
                    self._permission_warned = True
            except Exception as e:
                # For other Crazyradio exceptions (for instance if it's not attached)
                # ignore and keep scanning other link drivers.
                logger.warning(e)

        self._scan(address)

    def _display_input_device_error(self, error):
        self.cf.close_link()
        QMessageBox.critical(self, "Input device error", error)

    def _mux_selected(self, checked):
        """Called when a new mux is selected. The menu item contains a
        reference to the raw mux object as well as to the associated device
        sub-nodes"""
        if not checked:
            (mux, sub_nodes) = self.sender().data()
            for s in sub_nodes:
                s.setEnabled(False)
        else:
            (mux, sub_nodes) = self.sender().data()
            for s in sub_nodes:
                s.setEnabled(True)
            self.joystickReader.set_mux(mux=mux)

            # Go though the tree and select devices/mapping that was
            # selected before it was disabled.
            for role_node in sub_nodes:
                for dev_node in role_node.children():
                    if type(dev_node) is QAction and dev_node.isChecked():
                        dev_node.toggled.emit(True)

            self._update_input_device_footer()

    def _get_dev_status(self, device):
        msg = "{}".format(device.name)
        if device.supports_mapping:
            map_name = "No input mapping"
            if device.input_map:
                map_name = device.input_map_name
            msg += " ({})".format(map_name)
        return msg

    def _update_input_device_footer(self):
        """Update the footer in the bottom of the UI with status for the
        input device and its mapping"""

        msg = ""

        if len(self.joystickReader.available_devices()) > 0:
            mux = self.joystickReader._selected_mux
            msg = "Using {} mux with ".format(mux.name)
            for key in list(mux._devs.keys())[:-1]:
                if mux._devs[key]:
                    msg += "{}, ".format(self._get_dev_status(mux._devs[key]))
                else:
                    msg += "N/A, "
            # Last item
            key = list(mux._devs.keys())[-1]
            if mux._devs[key]:
                msg += "{}".format(self._get_dev_status(mux._devs[key]))
            else:
                msg += "N/A"
        else:
            msg = "No input device found"
        self._statusbar_label.setText(msg)

    def _inputdevice_selected(self, checked):
        """Called when a new input device has been selected from the menu. The
        data in the menu object is the associated map menu (directly under the
        item in the menu) and the raw device"""
        (map_menu, device, mux_menu) = self.sender().data()
        if not checked:
            if map_menu:
                map_menu.setEnabled(False)
                # Do not close the device, since we don't know exactly
                # how many devices the mux can have open. When selecting a
                # new mux the old one will take care of this.
        else:
            if map_menu:
                map_menu.setEnabled(True)

            (mux, sub_nodes) = mux_menu.data()
            for role_node in sub_nodes:
                for dev_node in role_node.children():
                    if type(dev_node) is QAction and dev_node.isChecked():
                        if device.id == dev_node.data()[1].id \
                                and dev_node is not self.sender():
                            dev_node.setChecked(False)

            role_in_mux = str(self.sender().parent().title()).strip()
            logger.info("Role of {} is {}".format(device.name,
                                                  role_in_mux))

            Config().set("input_device", str(device.name))

            self._mapping_support = self.joystickReader.start_input(
                device.name,
                role_in_mux)
        self._update_input_device_footer()

    def _inputconfig_selected(self, checked):
        """Called when a new configuration has been selected from the menu. The
        data in the menu object is a reference to the device QAction in parent
        menu. This contains a reference to the raw device."""
        if not checked:
            return

        selected_mapping = str(self.sender().text())
        device = self.sender().data().data()[1]
        self.joystickReader.set_input_map(device.name, selected_mapping)
        self._update_input_device_footer()

    def device_discovery(self, devs):
        """Called when new devices have been added"""
        for menu in self._all_role_menus:
            role_menu = menu["rolemenu"]
            mux_menu = menu["muxmenu"]
            dev_group = QActionGroup(role_menu)
            dev_group.setExclusive(True)
            for d in devs:
                dev_node = QAction(d.name, role_menu, checkable=True,
                                   enabled=True)
                role_menu.addAction(dev_node)
                dev_group.addAction(dev_node)
                dev_node.toggled.connect(self._inputdevice_selected)

                map_node = None
                if d.supports_mapping:
                    map_node = QMenu("    Input map", role_menu, enabled=False)
                    map_group = QActionGroup(role_menu)
                    map_group.setExclusive(True)
                    # Connect device node to map node for easy
                    # enabling/disabling when selection changes and device
                    # to easily enable it
                    dev_node.setData((map_node, d))
                    for c in ConfigManager().get_list_of_configs():
                        node = QAction(c, map_node, checkable=True,
                                       enabled=True)
                        node.toggled.connect(self._inputconfig_selected)
                        map_node.addAction(node)
                        # Connect all the map nodes back to the device
                        # action node where we can access the raw device
                        node.setData(dev_node)
                        map_group.addAction(node)
                        # If this device hasn't been found before, then
                        # select the default mapping for it.
                        if d not in self._available_devices:
                            last_map = Config().get("device_config_mapping")
                            if d.name in last_map and last_map[d.name] == c:
                                node.setChecked(True)
                    role_menu.addMenu(map_node)
                dev_node.setData((map_node, d, mux_menu))

        # Update the list of what devices we found
        # to avoid selecting default mapping for all devices when
        # a new one is inserted
        self._available_devices = ()
        for d in devs:
            self._available_devices += (d,)

        # Only enable MUX nodes if we have enough devies to cover
        # the roles
        for mux_node in self._all_mux_nodes:
            (mux, sub_nodes) = mux_node.data()
            if len(mux.supported_roles()) <= len(self._available_devices):
                mux_node.setEnabled(True)

        # TODO: Currently only supports selecting default mux
        if self._all_mux_nodes[0].isEnabled():
            self._all_mux_nodes[0].setChecked(True)

        # If the previous length of the available devies was 0, then select
        # the default on. If that's not available then select the first
        # on in the list.
        # TODO: This will only work for the "Normal" mux so this will be
        #       selected by default
        if Config().get("input_device") in [d.name for d in devs]:
            for dev_menu in self._all_role_menus[0]["rolemenu"].actions():
                if dev_menu.text() == Config().get("input_device"):
                    dev_menu.setChecked(True)
        else:
            # Select the first device in the first mux (will always be "Normal"
            # mux)
            self._all_role_menus[0]["rolemenu"].actions()[0].setChecked(True)
            logger.info("Select first device")

        self._update_input_device_footer()

    def _open_config_folder(self):
        QDesktopServices.openUrl(
            QUrl("file:///" +
                 QDir.toNativeSeparators(cfclient.config_path)))

    def closeAppRequest(self):
        self.close()
        sys.exit(0)


class ScannerThread(QThread):

    scanSignal = pyqtSignal(object)
    interfaceFoundSignal = pyqtSignal(object)

    def __init__(self):
        QThread.__init__(self)
        self.moveToThread(self)
        self.scanSignal.connect(self.scan)

    def scan(self, address):
        self.interfaceFoundSignal.emit(cflib.crtp.scan_interfaces(address))



# -*- coding: utf-8 -*-
#
#     ||          ____  _ __
#  +------+      / __ )(_) /_______________ _____  ___
#  | 0xBC |     / __  / / __/ ___/ ___/ __ `/_  / / _ \
#  +------+    / /_/ / / /_/ /__/ /  / /_/ / / /_/  __/
#   ||  ||    /_____/_/\__/\___/_/   \__,_/ /___/\___/
#
#  Copyright (C) 2021-2023 Bitcraze AB
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA  02110-1301, USA.

"""
Dialog box used to change lighthouse system type. Used from the lighthouse tab.
"""
import logging

import cfclient
from PyQt6 import QtWidgets
from PyQt6 import uic

__author__ = 'Bitcraze AB'
__all__ = ['LighthouseSystemTypeDialog']

logger = logging.getLogger(__name__)

(lighthouse_system_widget_class, connect_widget_base_class) = (
    uic.loadUiType(
        cfclient.module_path + '/ui/dialogs/lighthouse_system_type_dialog.ui')
)


class LighthouseSystemTypeDialog(QtWidgets.QWidget, lighthouse_system_widget_class):

    PARAM_GROUP = 'lighthouse'
    PARAM_NAME = 'systemType'

    VALUE_V1 = 1
    VALUE_V2 = 2

    def __init__(self, helper, *args):
        super(LighthouseSystemTypeDialog, self).__init__(*args)
        self.setupUi(self)

        self._helper = helper

        self._close_button.clicked.connect(self.close)

        self._radio_btn_v1.toggled.connect(self._type_toggled)
        self._radio_btn_v2.toggled.connect(self._type_toggled)

        self._curr_type = 0

    def get_system_type(self):
        system_type = self.VALUE_V2

        values = self._helper.cf.param.values
        if self.PARAM_NAME in values[self.PARAM_GROUP]:
            system_type = int(values[self.PARAM_GROUP][self.PARAM_NAME])

        return system_type

    def showEvent(self, event):
        self._curr_type = self.get_system_type()

        if self._curr_type == self.VALUE_V1:
            self._radio_btn_v1.setChecked(True)
        elif self._curr_type == self.VALUE_V2:
            self._radio_btn_v2.setChecked(True)

    def _type_toggled(self, *args):
        new_type = self.VALUE_V2

        if self._radio_btn_v1.isChecked():
            new_type = self.VALUE_V1
        elif self._radio_btn_v2.isChecked():
            new_type = self.VALUE_V2

        if new_type != self._curr_type:
            self._curr_type = new_type
            self._helper.cf.param.set_value(self.PARAM_GROUP + '.' + self.PARAM_NAME, self._curr_type)
